import argparse
import json
import random
import os
import time
import shutil
from typing import List
import warnings
warnings.filterwarnings('ignore')
import torch
import torch.nn as nn
import numpy as np
from gensim.models import Word2Vec
import shutil
from .src.finetuning import (
    run_finetuning_wkfl2,
    run_finetuning_wkfl3,
    setup_finetuning_input,
)
from .src.pretraining import run_pretraining, setup_pretraining_input
from .src.processing.context_generator import ContextGenerator
from .src.processing.generic_attributed_graph import GenericGraph
from .src.utils.data_utils import load_pretrained_node2vec
from .src.utils.evaluation import run_evaluation_main
from .src.utils.link_predict import find_optimal_cutoff, link_prediction_eval
from .src.utils.utils import get_id_map, load_pickle
torch.set_num_threads(50)

def get_set_embeddings_details(args):
    if not args.pretrained_embeddings:
        if args.pretrained_method == "compgcn":
            args.pretrained_embeddings = (
                f"{args.emb_dir}/{args.data_name}/"
                f"act_{args.data_name}_{args.node_edge_composition_func}_500.out"
            )
        elif args.pretrained_method == "node2vec":
            args.base_embedding_dim = 128
            args.pretrained_embeddings = (
                f"{args.emb_dir}/{args.data_name}/{args.data_name}.emd"
            )
    else:
        if args.pretrained_method == "compgcn":
            args.base_embedding_dim = 200
        elif args.pretrained_method == "node2vec":
            args.base_embedding_dim = 128
    return args.pretrained_embeddings, args.base_embedding_dim


def get_graph(args, data_path, false_edge_gen):
    print("\n Loading graph...")
    attr_graph = GenericGraph(data_path, false_edge_gen)
    context_gen = ContextGenerator(attr_graph, args.num_walks_per_node)

    return attr_graph, context_gen


def get_test_edges(paths: List[str], sep: str):
    # edges = set()
    edges = []
    for path in paths:
        with open(path, "r") as f:
            for line in f:
                tokens = line.strip().split(sep)
                etype = tokens[0]
                source = tokens[1]
                destination = tokens[2]
                label = tokens[3]
                edge = (etype, source, destination, label)
                # edges.add(edge)
                edges.append(edge)

    return edges

def Train(data_name='Sst_Sstr2',data_path='./Test/single-cell/data/',
    outdir='./Test/single-cell/data/Sst_Sstr2/output',
    pretrained_embeddings='./data_pca.emd',
    n_epochs = 50,ft_n_epochs=10):

    #torch.cuda.empty_cache()
    #os.environ["CUDA_VISIBLE_DEVICES"] = "0"
    #os.environ["CUDA_LAUNCH_BLOCKING"]="1"

    #torch.manual_seed(1234)
    parser = argparse.ArgumentParser()
    parser.add_argument("--device", default="cuda:0", type=str, help='specify device')
    parser.add_argument("--data_name", default="EFNA5_EPHA2", help="name of the dataset")
    parser.add_argument("--data_path", default="/mnt/test/a3/DeepSpa/dataset/PDAC/data/", help="path to dataset")
    parser.add_argument("--outdir", default="/mnt/test/a3/DeepSpa/dataset/PDAC/data/EFNA5_EPHA2/output", help="path to output dir")
    parser.add_argument(
        "--pretrained_embeddings", default="/mnt/test/a3/DeepSpa/dataset/PDAC/data/EFNA5_EPHA2/data_pca.emd",help="absolute path to pretrained embeddings"
    )
    parser.add_argument(
        "--pretrained_method", default="node2vec", help="compgcn|node2vec"
    )
    # Walks options
    parser.add_argument(
        "--beam_width",
        default=4,
        type=int,
        help="beam width used for generating random walks",
    )
    parser.add_argument(
        "--num_walks_per_node", default=8, type=int, help="walks per node"
    )
    parser.add_argument("--walk_type", default="dfs", help="walk type bfs/dfs")
    parser.add_argument("--max_length", default=5, type=int, help="max length of walks")
    parser.add_argument(
        "--n_pred", default=1, help="number of tokens masked to be predicted"
    )
    parser.add_argument(
        "--max_pred", default=1, help="max number of tokens masked to be predicted"
    )
    # Pretraining options
    parser.add_argument("--lr", default=0.0001, type=float, help="learning rate")
    parser.add_argument(
        "--n_epochs", default=50, type=int, help="number of epochs for training"
    )
    parser.add_argument(
        "--checkpoint", default=100, type=int, help="checkpoint for validation"
    )
    parser.add_argument(
        "--base_embedding_dim",
        default=200,
        type=int,
        help="dimension of base embedding",
    )
    parser.add_argument(
        "--batch_size",
        default=1024,
        type=int,
        help="number of data sample in each batch",
    )
    parser.add_argument(
        "--emb_dir",
        default="data",
        type=str,
        help="Used to generate embeddings path if --pretrained_embeddings is not set",
    )
    parser.add_argument(
        "--get_bert_encoder_embeddings",
        default=False,
        help="indicate if need to get node vectors from BERT encoder output, save code "
             "commented out in src/pretraining.py",
    )
    # BERT Layer options
    parser.add_argument(
        "--n_layers", default=4, type=int, help="number of encoder layers in bert"
    )
    parser.add_argument(
        "--d_model", default=200, type=int, help="embedding size in bert"
    )
    parser.add_argument("--d_k", default=64, type=int, help="dimension of K(=Q), V")
    parser.add_argument("--d_v", default=64, type=int, help="dimension of K(=Q), V")
    parser.add_argument("--n_heads", default=4, type=int, help="number of head in bert")
    parser.add_argument(
        "--d_ff",
        default=200 * 4,
        type=int,
        help="4*d_model, FeedForward dimension in bert",
    )
    # GCN Layer options
    parser.add_argument(
        "--is_pre_trained",default=False,
        action="store_true",
        help="if there is pretrained node embeddings",
    )
    parser.add_argument(
        "--gcn_option",
        default="no_gcn",
        help="preprocess bert input once or alternate gcn and bert, preprocess|alternate|no_gcn",
    )
    parser.add_argument(
        "--num_gcn_layers", default=4, type=int, help="number of gcn layers before bert"
    )
    parser.add_argument(
        "--node_edge_composition_func",
        default="no_rel",
        help="options for node and edge compostion, sub|circ_conv|mult|no_rel",
    )

    # Finetuning options
    parser.add_argument("--ft_lr", default=0.0001, type=float, help="learning rate")
    parser.add_argument(
        "--ft_batch_size",
        default=1024,
        type=int,
        help="number of data sample in each batch",
    )
    parser.add_argument(
        "--ft_checkpoint", default=1000, type=int, help="checkpoint for validation"
    )
    parser.add_argument(
        "--ft_d_ff", default=512, type=int, help="feedforward dimension in finetuning"
    )
    parser.add_argument(
        "--ft_layer", default="ffn", help="options for finetune layer: linear|ffn"
    )
    parser.add_argument(
        "--ft_drop_rate", default=0.1, type=float, help="dropout rate in finetuning"
    )
    parser.add_argument(
        "--ft_input_option",
        default="last4_cat",
        help="which output layer from graphbert will be used for finetuning, last|last4_cat|last4_sum",
    )
    parser.add_argument(
        "--false_edge_gen",
        default="basic",
        help="false edge generation pattern/double/basic",
    )
    parser.add_argument(
        "--ft_n_epochs", default=10, type=int, help="number of epochs for training"
    )
    parser.add_argument(
        "--path_option",
        default="random",
        help="fine tuning context generation: shortest/all/pattern/random",
    )
    args = parser.parse_args(args=[])

    # Default values
    args.pretrained_embeddings, args.base_embedding_dim = get_set_embeddings_details(
        args
    )
    args.d_model = args.base_embedding_dim
    args.d_ff = args.base_embedding_dim
    args.data_name = data_name
    args.data_path = data_path
    args.outdir = outdir
    args.pretrained_embeddings = pretrained_embeddings
    args.n_epochs = n_epochs
    args.ft_n_epochs = ft_n_epochs
    
    #print("Args ", str(args))

    data_path = f"{args.data_path}/{args.data_name}"
    attr_graph, context_gen = get_graph(args, data_path, args.false_edge_gen)
    attr_graph.dump_stats()
    print(attr_graph.G)
    stime = time.time()
    id_maps_dir = data_path
    ent2id = get_id_map(f"{id_maps_dir}/ent2id.txt")
    rel2id = get_id_map(f"{id_maps_dir}/rel2id.txt")
    print(len(ent2id), len(rel2id))

    # Load pretrained embedding from CompGCN
    if args.pretrained_method == "compgcn":
        pretrained_node_embedding = load_pickle(args.pretrained_embeddings)
    elif args.pretrained_method == "node2vec":
        pretrained_node_embedding = load_pretrained_node2vec(
            args.pretrained_embeddings, ent2id, args.base_embedding_dim
        )

    print(
        "No. of nodes with pretrained embedding: ",
        len(pretrained_node_embedding),
    )

    valid_path = data_path + "/valid.txt"
    valid_edges_paths = [valid_path]
    valid_edges = list(get_test_edges(valid_edges_paths, " "))
    test_path = data_path + "/test.txt"
    test_edges_paths = [test_path]
    test_edges = list(get_test_edges(test_edges_paths, " "))
    print("No. edges in test data: ", len(test_edges))
    '''
    if args.is_pre_trained:
        link_prediction_eval(
            valid_edges, test_edges, ent2id, rel2id, pretrained_node_embedding
        )
    else:
        print("No pretrained embedding, no need to evaluate workflow 1.\n")
    '''
    outdir = data_path+"/outdir"
    isExists = os.path.exists(outdir)
    if isExists:
        shutil.rmtree(outdir)
        os.makedirs(outdir)
    if not isExists:
        os.makedirs(outdir)


    print("***************PRETRAINING***************")
    pre_num_batches = setup_pretraining_input(args, attr_graph, context_gen, outdir)
    print("\n Run model for pre-training ...")
    
    # Masked nodes prediction
    pred_data, true_data = run_pretraining(
        args, attr_graph, pre_num_batches, pretrained_node_embedding, ent2id, rel2id
    )
    #print("\n Begin evaluation for node prediction...")
    # accu, mse = run_evaluation(pred_data, true_data)
    # Link prediction
    ft_num_batches = setup_finetuning_input(args, attr_graph, context_gen)
    print("\n !!!!")

    pred_data, true_data = run_finetuning_wkfl2(
        args, attr_graph, ft_num_batches, pretrained_node_embedding, ent2id, rel2id
    )
    print("\n Begin evaluation for link prediction...")
    valid_true_data = np.array(true_data["valid"])
    #print(valid_true_data)
    #threshold = find_optimal_cutoff(valid_true_data, pred_data["valid"])[0]
    threshold = 0.5
    #print(threshold)
    run_evaluation_main(
        test_edges, pred_data["test"], true_data["test"], threshold, header="workflow2"
    )
    # evaluate_all_epochs(args, attr_graph, ft_num_batches,
    #                    pretrained_node_embedding, ent2id, rel2id, test_edges)
    
    # WORKFLOW_3
    print("***************FINETUNING***************")
    print("\n Run model for finetuning ...")
    (
        pred_data_test,
        true_data_test,
        pred_data_valid,
        true_data_valid,
    ) = run_finetuning_wkfl3(
        args, attr_graph, ft_num_batches, pretrained_node_embedding, ent2id, rel2id
    )
    print("\n Begin evaluation for link prediction...")
    valid_true_data = np.array(true_data_valid)
    #threshold = 0.5

    #threshold = find_optimal_cutoff(valid_true_data, pred_data_valid)[0]
    #save the threshold values for later use
    #json.dump(threshold, open(args.outdir + "/thresholds.json", "w"))
    y_true,y_scores = run_evaluation_main(
        test_edges, pred_data_test, true_data_test, threshold, header="workflow3"
    )
    #np.save('./out/EFNA5_EPHA2/y_true.npy', y_true)
    #np.save('./out/EFNA5_EPHA2/y_pred.npy', y_scores)
    # evaluate after context inference

    etime = time.time()
    elapsed = etime - stime
    print(f"running time(seconds) on {args.data_name} data: {elapsed}")




'''
if __name__ == '__main__':


    Train(data_name='Sst_Sstr2',data_path='./Test/single-cell/data/',
    outdir='./Test/single-cell/data/Sst_Sstr2/output',
    pretrained_embeddings='./Test/single-cell/data/data_pca.emd')
'''


import torch
from torch.autograd import Variable

from .input_encoder import NodeEncoder


def get_norm_id(id_map, some_id):
    if some_id not in id_map:
        id_map[some_id] = len(id_map)
    return id_map[some_id]


def norm_graph(node_id_map, edge_id_map, edge_list):
    norm_edge_list = []
    for e in edge_list:
        norm_edge_list.append(
            (
                get_norm_id(node_id_map, e[0]),
                get_norm_id(node_id_map, e[1]),
                get_norm_id(edge_id_map, e[2]),
            )
        )
    return norm_edge_list


def norm_context_subgraphs_for_gcn_encoding(subgraph_list):
    """
    Function to transform a subgraph_list (same data structure passed to
    GraphBERT.forward() to a tuple of data structures that will be fed
    into GCN.
    Purpose of this data structure: Given a node embedding tensor from GCN
    or BERT, we want to know which node in a subgraph it corresponds to.

    param subgraph_list: A list of subgraphs of len batch_sz. Each subgraph
                         is a list of triples of form (src, dst, edge_type)
    ret norm_subgraph_list: A transformed version of subgraph_list where
                         each node id is normalized. See picture below for
                         example. Nodes mentioned in () and relations in [].
                               (0)     (5)              (0)     (2)
                                 \     /                  \     /
                                 [0]  [2]                 [0]  [1]
                                   \ /                      \ /
                                   (2)                      (1)
                              subgraph_list[i]      norm_subgraph_list[i]
    ret node_idx:        A (batch_sz, max_nodes) tensor. node_idx[i]
                         would return a tensor containing normalized node
                         ids for subgraph subgraph_list[i].
                         max_nodes is the maximum number of nodes present
                         in any of the subgraphs in subgraph_list.
                         If subgraph_list[i] has N_i nodes, then
                         node_idx[i][0:(N_i-1)] will return corresponding
                         node_ids in subgraph_list[i].
    ret edge_type_idx:   Similar (batch_sz, max_edge_types) tensor that
                         maps edge types found in subgraph_list[i] to a
                         range (0, num_unique_relations(subgraph_list[i]-1))
    ret batch_counts:    A list of pairs containing the number of nodes
                         and edges for each entry in subgraph_list
    """
    num_subgraphs = len(subgraph_list)
    norm_subgraph_list = []
    batch_counts = []
    batch_id_maps = []

    for i in range(num_subgraphs):
        node_id_map = dict()
        edge_type_map = dict()
        norm_subgraph_list.append(
            norm_graph(node_id_map, edge_type_map, subgraph_list[i])
        )
        batch_id_maps.append((node_id_map, edge_type_map))
        batch_counts.append((len(node_id_map), len(edge_type_map)))

    max_nodes = max(map(lambda counts: counts[0], batch_counts))
    max_relations = max(map(lambda counts: counts[1], batch_counts))
    node_idx = [[None for y in range(max_nodes)] for x in range(num_subgraphs)]
    # node_idx = torch.empty(num_subgraphs, max_nodes)
    edge_idx = torch.empty(num_subgraphs, max_relations)

    for i in range(num_subgraphs):
        node_id_map = batch_id_maps[i][0]
        edge_type_map = batch_id_maps[i][1]
        for orig_node_id in sorted(node_id_map.keys()):
            node_idx[i][node_id_map[orig_node_id]] = orig_node_id
        for orig_edge_type in sorted(edge_type_map.keys()):
            edge_idx[i][edge_type_map[orig_edge_type]] = orig_edge_type

    return norm_subgraph_list, node_idx, edge_idx, batch_counts


class GCNGraphEncoder(torch.nn.Module):
    def __init__(
        self,
        device,
        G,
        pretrained_node_embedding_tensor,
        is_pre_trained,
        base_embedding_dim,
        max_length,
        ent2id,
        rel2id,
    ):

        super().__init__()
        self.attr_graph = G
        self.base_embedding_dim = base_embedding_dim
        self.max_length = max_length
        self.ent2id = ent2id
        self.rel2id = rel2id
        self.no_nodes = self.attr_graph.get_number_of_nodes()
        self.no_relations = self.attr_graph.get_number_of_relations()
        # print('check *************', self.no_relations)

        self.node_embedding = NodeEncoder(
            base_embedding_dim,
            self.no_nodes,
            pretrained_node_embedding_tensor,
            is_pre_trained,device
        )
        self.relation_embed = torch.nn.Embedding(self.no_relations, base_embedding_dim)
        # updated for KGAT data
        # torch.nn.Embedding(len(self.rel2id), base_embedding_dim) 
        self.relation_embed.weight.data.uniform_(-1, 1)

        self.special_tokens = {"[PAD]": 0, "[MASK]": 1}
        self.special_embed = torch.nn.Embedding(
            len(self.special_tokens), base_embedding_dim
        )
        self.special_embed.weight.data.uniform_(-1, 1)

    def forward(self, subgraphs_list, masked_nodes):
        num_subgraphs = len(subgraphs_list)
        norm_subgraph_list = []
        batch_counts = []
        batch_id_maps = []

        for ii, _ in enumerate(subgraphs_list):
            node_id_map = dict()
            edge_type_map = dict()
            norm_subgraph_list.append(
                norm_graph(node_id_map, edge_type_map, subgraphs_list[ii])
            )
            batch_id_maps.append((node_id_map, edge_type_map))
            batch_counts.append((len(node_id_map), len(edge_type_map)))

        # max_nodes = max(map(lambda counts: counts[0], batch_counts))
        max_relations = max(map(lambda counts: counts[1], batch_counts))

        node_emb = torch.zeros(
            num_subgraphs, self.max_length + 1, self.base_embedding_dim
        )
        #print(node_emb.shape)
        relation_emb = torch.zeros(
            num_subgraphs, max_relations, self.base_embedding_dim
        )
        #print(self.node_embedding(2397))
        for ii in range(len(subgraphs_list)):
            node_id_map = batch_id_maps[ii][0]
            edge_type_map = batch_id_maps[ii][1]
            masked_set = masked_nodes[ii].cpu().numpy().tolist()

            for node_id, norm_node_id in node_id_map.items():
                if node_id not in masked_set:  # used to ignore the masked nodes
                    try:
                        normalized_node_id = self.ent2id[node_id]
                    except KeyError:  # FIXME - replaced bare except
                        normalized_node_id = int(node_id)
                    #print(normalized_node_id)

                    node_emb[ii][norm_node_id] = self.node_embedding(normalized_node_id)

            # print('edge_type_map\n', edge_type_map)
            for relation_id, norm_relation_id in edge_type_map.items():
                try:
                    tmp = Variable(torch.LongTensor([int(self.rel2id[relation_id])]))
                except KeyError:  # FIXME - replaced bare except
                    tmp = Variable(torch.LongTensor([int(relation_id) - 1]))

                relation_emb[ii][norm_relation_id] = self.relation_embed(tmp)

        # get embeddings for special tokens
        # will be used for masking and padding before bert layer
        special_tokens_embed = {}
        for token in self.special_tokens:
            node_id = Variable(torch.LongTensor([self.special_tokens[token]]))
            tmp_embed = self.special_embed(node_id)
            special_tokens_embed[self.special_tokens[token] + self.no_nodes] = {
                "token": token,
                "embed": tmp_embed,
            }

        return (
            norm_subgraph_list,
            node_emb,
            relation_emb,
            batch_id_maps,
            special_tokens_embed,
        )



name = "DeepTalk_ST"

import warnings
warnings.filterwarnings('ignore')

from .utils import *
from .map_utils import *
from .plot_utils import *
from .decon_utils import *
from .cci_vis import *
from .ccc_predict import *


from ..helper import *
from .compgcn_conv import CompGCNConv
from .compgcn_conv_basis import CompGCNConvBasis

class BaseModel(torch.nn.Module):
	def __init__(self, params):
		super(BaseModel, self).__init__()

		self.p		= params
		self.bceloss	= torch.nn.BCELoss()

	def loss(self, pred, true_label):
		return self.bceloss(pred, true_label)

class CompGCNBase(BaseModel):
	def __init__(self, edge_index, edge_type, num_rel, params=None):
		super(CompGCNBase, self).__init__(params)

		self.edge_index		= edge_index
		self.edge_type		= edge_type
		self.p.gcn_dim		= self.p.embed_dim if self.p.gcn_layer == 1 else self.p.gcn_dim
		self.init_embed		= get_param((self.p.num_ent,   self.p.init_dim))
		self.device		= self.edge_index.device

		if self.p.num_bases > 0:
			self.init_rel  = get_param((self.p.num_bases,   self.p.init_dim))
		else:
			if self.p.score_func == 'transe': 	self.init_rel = get_param((num_rel,   self.p.init_dim))
			else: 					self.init_rel = get_param((num_rel*2, self.p.init_dim))

		if self.p.num_bases > 0:
			self.conv1 = CompGCNConvBasis(self.p.init_dim, self.p.gcn_dim, num_rel, self.p.num_bases, act=self.act, params=self.p)
			self.conv2 = CompGCNConv(self.p.gcn_dim,    self.p.embed_dim, self.p.opn) if self.p.gcn_layer == 2 else None
		else:
			self.conv1 = CompGCNConv(self.p.init_dim, self.p.gcn_dim,self.p.opn)
			self.conv2 = CompGCNConv(self.p.gcn_dim,    self.p.embed_dim,self.p.opn) if self.p.gcn_layer == 2 else None

		self.register_parameter('bias', Parameter(torch.zeros(self.p.num_ent)))

	def forward_base(self, sub, rel, drop1, drop2):

		r	= self.init_rel if self.p.score_func != 'transe' else torch.cat([self.init_rel, -self.init_rel], dim=0)
		x, r	= self.conv1(self.init_embed, self.edge_index, self.edge_type, rel_embed=r)
		x	= drop1(x)
		x, r	= self.conv2(x, self.edge_index, self.edge_type, rel_embed=r) 	if self.p.gcn_layer == 2 else (x, r)
		x	= drop2(x) 							if self.p.gcn_layer == 2 else x

		sub_emb	= torch.index_select(x, 0, sub)
		rel_emb	= torch.index_select(r, 0, rel)

		return sub_emb, rel_emb, x


class CompGCN_TransE(CompGCNBase):
	def __init__(self, edge_index, edge_type, params=None):
		super(self.__class__, self).__init__(edge_index, edge_type, params.num_rel, params)
		self.drop = torch.nn.Dropout(self.p.hid_drop)

	def forward(self, sub, rel):

		sub_emb, rel_emb, all_ent	= self.forward_base(sub, rel, self.drop, self.drop)
		obj_emb				= sub_emb + rel_emb

		x	= self.p.gamma - torch.norm(obj_emb.unsqueeze(1) - all_ent, p=1, dim=2)		
		score	= torch.sigmoid(x)

		return score

class CompGCN_DistMult(CompGCNBase):
	def __init__(self, edge_index, edge_type, params=None):
		super(self.__class__, self).__init__(edge_index, edge_type, params.num_rel, params)
		self.drop = torch.nn.Dropout(self.p.hid_drop)

	def forward(self, sub, rel):

		sub_emb, rel_emb, all_ent	= self.forward_base(sub, rel, self.drop, self.drop)
		obj_emb				= sub_emb * rel_emb

		x = torch.mm(obj_emb, all_ent.transpose(1, 0))
		x += self.bias.expand_as(x)

		score = torch.sigmoid(x)
		return score

class CompGCN_ConvE(CompGCNBase):
	def __init__(self, edge_index, edge_type, params=None):
		super(self.__class__, self).__init__(edge_index, edge_type, params.num_rel, params)

		self.bn0		= torch.nn.BatchNorm2d(1)
		self.bn1		= torch.nn.BatchNorm2d(self.p.num_filt)
		self.bn2		= torch.nn.BatchNorm1d(self.p.embed_dim)
		
		self.hidden_drop	= torch.nn.Dropout(self.p.hid_drop)
		self.hidden_drop2	= torch.nn.Dropout(self.p.hid_drop2)
		self.feature_drop	= torch.nn.Dropout(self.p.feat_drop)
		self.m_conv1		= torch.nn.Conv2d(1, out_channels=self.p.num_filt, kernel_size=(self.p.ker_sz, self.p.ker_sz), stride=1, padding=0, bias=self.p.bias)

		flat_sz_h		= int(2*self.p.k_w) - self.p.ker_sz + 1
		flat_sz_w		= self.p.k_h 	    - self.p.ker_sz + 1
		self.flat_sz		= flat_sz_h*flat_sz_w*self.p.num_filt
		self.fc			= torch.nn.Linear(self.flat_sz, self.p.embed_dim)

	def concat(self, e1_embed, rel_embed):
		e1_embed	= e1_embed. view(-1, 1, self.p.embed_dim)
		rel_embed	= rel_embed.view(-1, 1, self.p.embed_dim)
		stack_inp	= torch.cat([e1_embed, rel_embed], 1)
		stack_inp	= torch.transpose(stack_inp, 2, 1).reshape((-1, 1, 2*self.p.k_w, self.p.k_h))
		return stack_inp

	def forward(self, sub, rel):

		sub_emb, rel_emb, all_ent	= self.forward_base(sub, rel, self.hidden_drop, self.feature_drop)
		stk_inp				= self.concat(sub_emb, rel_emb)
		x				= self.bn0(stk_inp)
		x				= self.m_conv1(x)
		x				= self.bn1(x)
		x				= F.relu(x)
		x				= self.feature_drop(x)
		x				= x.view(-1, self.flat_sz)
		x				= self.fc(x)
		x				= self.hidden_drop2(x)
		x				= self.bn2(x)
		x				= F.relu(x)

		x = torch.mm(x, all_ent.transpose(1,0))
		x += self.bias.expand_as(x)

		score = torch.sigmoid(x)
		return score



from copy import deepcopy
from pathlib import Path
import torch
from torch import nn
from scipy import spatial
import time
from torch.autograd import Variable
import torch.nn.functional as F
import numpy as np
def MLP(channels: list, do_bn=True):
    """ Multi-layer perceptron """
    n = len(channels)
    layers = []
    for i in range(1, n):
        layers.append(
            nn.Conv1d(channels[i - 1], channels[i], kernel_size=1, bias=True))
        if i < (n-1):
            if do_bn:
                # layers.append(nn.BatchNorm1d(channels[i]))
                layers.append(nn.InstanceNorm1d(channels[i]))
            layers.append(nn.ReLU())
    return nn.Sequential(*layers)


def normalize_keypoints(kpts, image_shape):
    """ Normalize keypoints locations based on image         print(desc0.shape)
image_shape"""
    #print(image_shape)
    #_, _, height, width = image_shape
    height, width = image_shape.numpy()[0],image_shape.numpy()[1]
    #height, width = image_shape.cpu().numpy()[0][0],image_shape.cpu().numpy()[0][1]
    one = kpts.new_tensor(1)
    size = torch.stack([one*width, one*height])[None]
    center = size / 2
    scaling = size.max(1, keepdim=True).values * 0.7
    return (kpts - center[:, None, :]) / scaling[:, None, :]

class KeypointEncoder(nn.Module):
    """ Joint encoding of visual appearance and location using MLPs"""
    def __init__(self, feature_dim, layers):
        super().__init__()
        self.encoder = MLP([2] + layers + [feature_dim])
        nn.init.constant_(self.encoder[-1].bias, 0.0)

    def forward(self, kpts):
        #inputs = [kpts.transpose(1, 2), scores.unsqueeze(1)]
        inputs = [kpts.transpose(1, 2)]
        return self.encoder(torch.cat(inputs, dim=1))


def attention(query, key, value):
    dim = query.shape[1]
    scores = torch.einsum('bdhn,bdhm->bhnm', query, key) / dim**.5
    prob = torch.nn.functional.softmax(scores, dim=-1)
    return torch.einsum('bhnm,bdhm->bdhn', prob, value), prob


class My_loss_new(nn.Module):
    def __init__(self):
        super().__init__()
        
    def forward(self, M, S, T):

        SSS = (torch.mm((M[0,:,:].softmax(dim=-1)).transpose(0,1),S[:,0,:].transpose(0,1)))
        T1 = T[:,0,:].transpose(0,1)

        CT_COR = F.cosine_similarity(SSS,T1,dim=0)
        CT_COR_mean0 = torch.mean((CT_COR))
        CT_COR_mean0 = torch.reshape(CT_COR_mean0, (1, -1))

        CT_COR1 = F.cosine_similarity(SSS,T1,dim=1)
        CT_COR_mean1 = torch.mean((CT_COR1))
        CT_COR_mean1 = torch.reshape(CT_COR_mean1, (1, -1))

        loss = - 1*CT_COR_mean0[0] +0*CT_COR_mean1[0]
        #loss = - CT_COR_mean[0]
        score0 = CT_COR_mean0[0]
        score1 = CT_COR_mean1[0]
        return loss, score0, score1

class MultiHeadedAttention(nn.Module):
    """ Multi-head attention to increase model expressivitiy """
    def __init__(self, num_heads: int, d_model: int):
        super().__init__()
        assert d_model % num_heads == 0
        self.dim = d_model // num_heads
        self.num_heads = num_heads
        self.merge = nn.Conv1d(d_model, d_model, kernel_size=1)
        self.proj = nn.ModuleList([deepcopy(self.merge) for _ in range(3)])

    def forward(self, query, key, value):
        batch_dim = query.size(0)
        query, key, value = [l(x).view(batch_dim, self.dim, self.num_heads, -1)
                             for l, x in zip(self.proj, (query, key, value))]
        x, prob = attention(query, key, value)
        self.prob.append(prob)
        return self.merge(x.contiguous().view(batch_dim, self.dim*self.num_heads, -1))


class AttentionalPropagation(nn.Module):
    def __init__(self, feature_dim: int, num_heads: int):
        super().__init__()
        self.attn = MultiHeadedAttention(num_heads, feature_dim)
        self.mlp = MLP([feature_dim*2, feature_dim*2, feature_dim])
        nn.init.constant_(self.mlp[-1].bias, 0.0)

    def forward(self, x, source):
        message = self.attn(x, source, source)
        return self.mlp(torch.cat([x, message], dim=1))


class AttentionalGNN(nn.Module):
    def __init__(self, feature_dim: int, layer_names: list):
        super().__init__()
        self.layers = nn.ModuleList([
            AttentionalPropagation(feature_dim, 1)
            for _ in range(len(layer_names))])
        self.names = layer_names

    def forward(self, desc0, desc1):
        for layer, name in zip(self.layers, self.names):
            layer.attn.prob = []
            if name == 'cross':
                src0, src1 = desc1, desc0
            else:  # if name == 'self':
                src0, src1 = desc0, desc1
            delta0, delta1 = layer(desc0, src0), layer(desc1, src1)
            desc0, desc1 = (desc0 + delta0), (desc1 + delta1)
        return desc0, desc1


def arange_like(x, dim: int):
    return x.new_ones(x.shape[dim]).cumsum(0) - 1  # traceable in 1.1


class Matchsc_st(nn.Module):

    default_config = {
        'keypoint_encoder': [8,16,32],
        'GNN_layers': ['self','cross'] * 1,
    }

    def __init__(self, config):
        super().__init__()
        self.config = {**self.default_config, **config}

        self.kenc = KeypointEncoder(
            self.config['descriptor_dim'], self.config['keypoint_encoder'])

        self.gnn = AttentionalGNN(
            self.config['descriptor_dim'], self.config['GNN_layers'])

        self.final_proj = nn.Conv1d(
            self.config['descriptor_dim'], self.config['descriptor_dim'],
            kernel_size=1, bias=True)

    def forward(self, data):
        """Run SuperGlue on a pair of keypoints and descriptors"""
        desc0, desc1 = data['descrip0Scale'], data['descrip1Scale']
        #desc0, desc1 = data['descriptors0'], data['descriptors1']

        kpts0, kpts1 = data['keypoints0'], data['keypoints1']

        desc0 = desc0.transpose(0,1)
        desc1 = desc1.transpose(0,1)

        kpts0 = torch.reshape(kpts0, (1, -1, 2))
        kpts1 = torch.reshape(kpts1, (1, -1, 2))

        # Keypoint normalization.
        kpts0 = normalize_keypoints(kpts0, data['sc_location'])
        kpts1 = normalize_keypoints(kpts1, data['st_location'])
        #print(data['sc_location'].shape)
        #print(kpts0.shape)
        # Keypoint MLP encoder.
        desc0 = desc0 + self.kenc(kpts0)
        desc1 = desc1 + self.kenc(kpts1)
        #print(desc1.shape)
        # Multi-layer Transformer network.
        desc0, desc1 = self.gnn(desc0, desc1)

        # Final MLP projection.
        #desc0, desc1 = self.final_proj(desc0), self.final_proj(desc1)
        scores = torch.einsum('bdn,bdm->bnm', desc0, desc1)
        #scores = scores / self.config['descriptor_dim']**.5
        scores = scores / self.config['descriptor_dim']

        return {
            'scores':scores
            }

    def get_loss(self,scores,data):

        SC = data['descriptors0'].float()
        ST = data['descriptors1'].float()

        loss_fun = My_loss_new()

        loss,score0,score1 = loss_fun(scores['scores'].float(), SC, ST)

        return loss,score0,score1







import pickle
import random

import networkx as nx
import numpy as np

from .negative_sampling import NegativeSampleGenerator


class AttributedGraph:
    def __init__(self):
        """
        Base class for all graph loaders.
        Assumes derived classes will create a networkx graph object of type
        1) nx.Graph() or nx.DiGraph()
        2) A node id to row id (0 to n-1) dictionary
        3) node_attr_dictionary per node type
        {"nodetype" : pd.dataframe}
        4) set of unique relations
        Note : The random walk methods implemented on top of networkx do not
        work with nx.MultiGraph() and hence MultiGraph object is not supported
        """
        self.G = nx.Graph()
        self.node_attr_dfs = dict()
        self.unique_relations = set()
        self.node_types = dict()
        self.normalized_node_id_map = dict()
        self.train_edges = list()
        self.valid_edges = list()
        self.test_edges = list()
        self.relation_to_id = dict()
        self.id_to_relation = dict()
        self.nodeid2rowid = dict()
        self.rowid2nodeid = dict()
        self.rowid2vocabid = dict()

    def load_graph(self, edgefile, node_attr_file=None):
        """
        This method should be implemented by all DataLoader classes to fill in
        values for
        1) self.G
        2) self.nodeid
        3) self.node_attrs
        4) self.unique_relations
        see SampleParser for example
        """
        return

    def get_normalized_node_id(self, node_id):
        # throw error?
        if str(node_id) not in self.G.nodes():
            print("Could not find node in graph", str(node_id))
            return None
        else:
            return self.normalized_node_id_map[str(node_id)]

    def create_normalized_node_id_map(self):
        cnt = 0
        for node_id in self.G.nodes():
            self.normalized_node_id_map[node_id] = cnt
            cnt += 1
        return

    def dump_graph(self, outdir, fr_valid_edges, fr_test_edges):
        print("Saving  graph at ", outdir)
        self.generate_link_prediction_dataset(outdir, fr_valid_edges, fr_test_edges)
        try:
            with open(outdir + "/normalized.txt", "w") as f:
                for u, v in self.G.edges():
                    relation = self.G.edges[u, v]["label"]
                    # f.write(str(self.normalized_node_id_map[u]) +
                    # " " + str(self.normalized_node_id_map[v]) + " " + relation + "\n")
                    f.write(f"{u}\t{v} {relation}\n")
        except IOError:
            print("Unable to write normalized graph at ", outdir)

        print("Saving test train validation set for link prediction")
        with open(outdir + "/train.txt", "w") as f:
            for edge in self.train_edges:
                f.write(f"{edge[0]} {edge[1]} {edge[2]} {edge[3]}\n")

        with open(outdir + "/valid.txt", "w") as f:
            for edge in self.valid_edges:
                f.write(f"{edge[0]} {edge[1]} {edge[2]} {edge[3]}\n")

        with open(outdir + "/test.txt", "w") as f:
            for edge in self.test_edges:
                f.write(f"{edge[0]} {edge[1]} {edge[2]} {edge[3]}\n")
        return

    def get_graph(self):
        return self.G

    def get_nodes(self):
        return self.G.nodes()

    def get_number_of_nodes(self):
        return len(self.G.nodes())

    def get_number_of_edges(self):
        return len(self.G.edges())

    def get_number_of_relations(self):
        return len(self.unique_relations)

    # def get_node_attr(self, node_id):
    #    return self.node_attrs[node_id]

    def get_edge_attr(self, u, v):
        return self.G.edges[u, v]["label"]

    def get_edges(self, node_id):  # FIXME, removed dangerous list init
        if len(node_id) == 0:
            return self.G.edges.data()
        else:
            return list(self.G.neighbors(node_id))

    def get_node_attr_df(self):
        return self.node_attr_dfs

    def get_node_types(self):
        return self.node_types

    def get_node_type(self, node_id):
        if node_id in self.node_types:
            return self.node_types[node_id]
        else:
            return ""

    def get_num_node_types(self):
        return len(self.node_attr_dfs)

    def get_node_attr(self, node_id, attribute_label=None):
        node_type = self.node_types[node_id]
        try:
            if attribute_label:
                return self.node_attr_dfs[node_type].at[node_id, attribute_label]
            else:
                return self.node_attr_dfs[node_type].loc[node_id]
        except KeyError:
            print("Could not find this node attribute", node_id, attribute_label)
            return None

    def get_random_non_edges(self, cnt):
        result = []
        nodes = list(self.G.nodes())
        relations = self.unique_relations
        print("Generating false edges", cnt)
        for _ in range(cnt):
            pair = random.sample(nodes, 2)
            u = pair[0]
            v = pair[1]
            if v not in list(self.G.neighbors(u)):
                result.append((random.sample(relations, 1)[0], u, v, "0"))

        return result

    def generate_link_prediction_dataset(self, outdir, fr_valid_edges, fr_test_edges):
        print("In Default AttributedGraph LinkPrediction generation")
        false_edge_gen = self.false_edge_gen
        all_positive_edges = self.map_positive_edges()
        num_edges = self.G.number_of_edges()
        print(num_edges)
        num_test_edges = int(num_edges * fr_test_edges)
        num_valid_edges = int(num_edges * fr_valid_edges)
        num_train_edges = num_edges - (num_valid_edges + num_test_edges)
        self.train_edges = all_positive_edges[0:num_train_edges]
        self.valid_edges = all_positive_edges[
            num_train_edges : num_train_edges + num_valid_edges
        ]
        self.test_edges = all_positive_edges[num_train_edges + num_valid_edges :]
        if false_edge_gen == "pattern":
            self.train_edges = self.generate_false_edges3(self.train_edges)
            self.valid_edges = self.generate_false_edges3(self.valid_edges)
            self.test_edges = self.generate_false_edges3(self.test_edges)
        elif false_edge_gen == "double":
            self.train_edges = self.generate_false_edges2(self.train_edges)
            self.valid_edges = self.generate_false_edges2(self.valid_edges)
            self.test_edges = self.generate_false_edges2(self.test_edges)
        else:
            self.train_edges = self.generate_false_edges(self.train_edges)
            self.valid_edges = self.generate_false_edges(self.valid_edges)
            self.test_edges = self.generate_false_edges(self.test_edges)

        self.valid_edges, self.test_edges = self.remove_coldstart_valid_test(
            self.train_edges, self.valid_edges, self.test_edges
        )
        # print('Dump link prediction edges ...')
        # self.dump_edges(outdir+ "/train.txt", self.train_edges)
        # self.dump_edges(outdir+ "/valid.txt", self.valid_edges)
        # self.dump_edges(outdir+ "/test.txt", self.test_edges)
        return self.train_edges, self.valid_edges, self.test_edges

    def remove_coldstart_valid_test(self, train_edges, valid_edges, test_edges):
        nodes_in_training_set = []
        for r, u, v, _ in train_edges:
            nodes_in_training_set.append(u)
            nodes_in_training_set.append(v)

        cleaned_valid_edges = []
        for r, u, v, label in valid_edges:
            if u in nodes_in_training_set and v in nodes_in_training_set:
                cleaned_valid_edges.append((r, u, v, label))

        cleaned_test_edges = []
        for r, u, v, label in test_edges:
            if u in nodes_in_training_set and v in nodes_in_training_set:
                cleaned_test_edges.append((r, u, v, label))
        return cleaned_valid_edges, cleaned_test_edges

    def generate_false_edges(self, positive_edge_list):
        """
        This method takes in as input a list of positive edges of form
        (relation, u, v, "1")
        and generates ann equal number of false edges (not present in graph),
        such that for each edge in positive_edge_list:
            (relation,u, v', "0"), here u->v' does not exist in self.G
        returns : random.shuffle(true_edges + false_edges)
        """
        count = len(positive_edge_list)
        # collect nodes of different types
        node_type_to_ids = dict()
        for node_id, node_type in self.node_types.items():
            if node_type in node_type_to_ids:
                node_type_to_ids[node_type].append(node_id)
            else:
                node_type_to_ids[node_type] = [node_id]

        false_edges = set()
        while len(false_edges) < count:
            for relation, source, target, _ in positive_edge_list:
                target_type = self.node_types[target]
                false_target = random.sample(node_type_to_ids[target_type], 1)[0]
                all_source_nbrs = list(set(list(nx.all_neighbors(self.G, source))))
                if false_target not in all_source_nbrs:
                    if len(false_edges) < count:
                        false_edges.add((relation, source, false_target, "0"))
                    else:
                        break

        final_edges = list(set(positive_edge_list + list(false_edges)))
        random.shuffle(final_edges)
        return final_edges
    

    def generate_false_edges1(self, positive_edge_list, false_edge_list):
        """
        This method takes in as input a list of positive edges of form
        (relation, u, v, "1")
        and generates ann equal number of false edges (not present in graph),
        such that for each edge in positive_edge_list:
            (relation,u, v', "0"), here u->v' does not exist in self.G
        returns : random.shuffle(true_edges + false_edges)
        """
        
        final_edges = list(set(positive_edge_list + false_edge_list))
        random.shuffle(final_edges)
        return final_edges


    def generate_false_edges2(self, positive_edge_list):
        """
        This method takes in as input a list of positive edges of form
        (relation, u, v, "1")
        and tries to generates double number of false edges (not present in graph),
        such that for each edge in positive_edge_list:
            (relation,u, v', "0"), here u -> v' does not exist in self.G
            (relation,u', v, "0"), here u'-> v does not exist in self.G
        returns : random.shuffle(true_edges + false_edges)
        """

        # collect nodes of different types
        node_type_to_ids = dict()
        for node_id, node_type in self.node_types.items():
            if node_type in node_type_to_ids:
                node_type_to_ids[node_type].append(node_id)
            else:
                node_type_to_ids[node_type] = [node_id]

        # generate false edges for every positive example
        false_edges = []
        for relation, source, target, _ in positive_edge_list:
            # generate false edges of type (source, relation, false_target) for every
            # (source, relation, target) in positive_edge_list
            target_type = self.node_types[target]
            false_target = random.sample(node_type_to_ids[target_type], 1)[0]
            all_source_nbrs = list(set(list(nx.all_neighbors(self.G, source))))
            if false_target not in all_source_nbrs:
                false_edges.append((relation, source, false_target, "0"))

            # generate false edges of type (false_source, relation, target) for every
            # (source, relation, target) in positive_edge_list
            source_type = self.node_types[source]
            false_source = random.sample(node_type_to_ids[source_type], 1)[0]
            all_target_nbrs = list(set(list(nx.all_neighbors(self.G, target))))
            if false_source not in all_target_nbrs:
                false_edges.append((relation, false_source, target, "0"))

        # generate false edges of type (false_source, relation, target) for every
        final_edges = list(set(positive_edge_list + false_edges))
        random.shuffle(final_edges)
        print(len(final_edges))
        return final_edges

    def generate_false_edges3(
        self, positive_edge_list, num_false_count_per_true_edge=5
    ):
        # generate negative edges of
        # num_false_count_per_true_edge*len(self.G.edges()) for graph object
        # 1) find valid edge types for each node
        # 2) group nodes that have same set of valid edges to determine node types
        # 3) for each node type pair
        #    (relation[i],u', v', "0"), here u'.type -> v'.type does not exist in self.G
        #    relation[i] = self.unique_relations[i]
        # returns : random.shuffle(true_edges + false_edges)
        # get {node -> set(valid edge types)}

        generator = NegativeSampleGenerator()
        false_edges = generator.gen_negative_triples(
            self.G, len(positive_edge_list) * num_false_count_per_true_edge
        )

        final_edges = list(positive_edge_list + false_edges)
        random.shuffle(final_edges)
        return final_edges

    def map_positive_edges(self):
        mappped_edge_list_by_type = []
        for u, v in self.G.edges():
            relation = self.G.edges[u, v]["label"]
            mappped_edge_list_by_type.append((relation, u, v, "1"))
        return mappped_edge_list_by_type

    def dump_edges(self, filename, edges, label=True):
        random.shuffle(edges)
        with open(filename, "w") as f:
            for edge in edges:
                if label:
                    f.write(f"{edge[0]} {edge[1]} {edge[2]} {edge[3]}\n")
                else:
                    f.write(f"{edge[0]} {edge[1]} {edge[2]}\n")
        return

    def dump_stats(self):
        print("DataSet Stats :")
        print("Number of Nodes", self.get_number_of_nodes())
        print("Number of Edges", self.get_number_of_edges())
        print("Number of Node types", self.get_num_node_types())
        print("Number of relations", self.get_number_of_relations())

    def get_nodeid2rowid(self):
        self.nodeid2rowid = {}
        for node_type in self.node_attr_dfs:
            df = self.node_attr_dfs[node_type]
            node_set = df.index.tolist()
            node2row = {}
            for i, _ in enumerate(node_set):
                node2row[node_set[i]] = i
            # print('\n nodeid2rowid: \n', node2row)
            self.nodeid2rowid[node_type] = node2row

        return self.nodeid2rowid

    def get_rowid2nodeid(self):
        self.rowid2nodeid = {}
        for node_type in self.nodeid2rowid:
            self.rowid2nodeid[node_type] = {}
            for itm in self.nodeid2rowid[node_type]:
                self.rowid2nodeid[node_type][self.nodeid2rowid[node_type][itm]] = itm

        return self.rowid2nodeid

    def get_rowid2vocabid_map(self):
        self.rowid2vocabid = {}
        for node_type in self.rowid2nodeid:
            self.rowid2vocabid[node_type] = {}
            for row_id in self.rowid2nodeid[node_type]:
                graph_node_id = self.rowid2nodeid[node_type][row_id]
                vocabid = self.get_normalized_node_id(graph_node_id)
                self.rowid2vocabid[node_type][row_id] = vocabid

        return self.rowid2vocabid

    def get_rowid2vocabid(self, row_id, node_type):
        # return an interger of vocab id
        return self.rowid2vocabid[node_type][int(row_id)]

    def get_rowid2graphid(self, row_id, node_type):
        try:
            return self.rowid2nodeid[node_type][int(row_id)]
        except KeyError:
            print("Node type and row id not found", row_id, node_type)
            return -1

    def get_graph_node_id(self, node_type_with_row_id):
        """take the string containing nodetype_noderowid."""
        arr = node_type_with_row_id.split("_")
        node_type = arr[0]
        row_id = arr[1]
        graph_node_id = self.rowid2nodeid[node_type][int(row_id)]
        return graph_node_id

    def set_relation_id(self):
        relations = list(self.unique_relations)
        self.relation_to_id = {relations[ii]: ii for ii in range(len(relations))}
        self.id_to_relation = {v: k for k, v in self.relation_to_id.items()}

    def get_relation_id(self, relation):
        try:
            return self.relation_to_id[relation]
        except KeyError:
            print("Could not find relation", relation)
            return -1

    def get_relation_from_id(self, relation_id):
        try:
            return self.id_to_relation[relation_id]
        except KeyError:
            print("Could not find relation", relation_id)
            return -1

    def dump_graph_transformer_networks_dataset(self, dump_dir):
        # Dump features pkl
        #   DBLP: len=334 int32 numpy ndarray per node
        #   ACM: len=1902 int64 numpy ndarray per node
        #   IMDB: len=1256 float32 numpy ndarray per node
        no_features = True
        node_features = []
        for _ in range(self.get_number_of_nodes()):
            # Currently we don't have features so I am just dumping an empty tensor
            if no_features:
                node_feature = np.zeros(shape=(100,))
            else:
                pass
            node_features.append(node_feature)
        node_features_path = f"{dump_dir}/node_features.pkl"
        with open(node_features_path, "wb") as f:
            pickle.dump(node_features, f)

        # Dump edges pkl
        #   All datasets from GTN have 4 relationship types

        # Dump labels pkl
        #   DBLP: 800, 400, 2857
        #   ACM: 600, 300, 2125
        #   IMDB: 300, 300, 2339



"""
This module includes plotting utility functions.
"""
import numpy as np
import matplotlib.pyplot as plt
import logging
import seaborn as sns
from scipy.stats import entropy
import scanpy as sc
from scipy.sparse.csc import csc_matrix
from scipy.sparse.csr import csr_matrix

from .utils import * 
from . import map_utils as mu
import pandas as pd
import logging
import matplotlib as mpl
from matplotlib.gridspec import GridSpec


def q_value(data, perc):
    """
    Computes min and max values according to percentile for colormap in plot functions
    Args:
        data (numpy array): input
        perc (float): percentile that between 0 and 100 inclusive
    Returns:
        tuple of floats: will be later used to define the data range covers by the colormap
    """
    vmin = np.nanpercentile(data, perc)
    vmax = np.nanpercentile(data, 100 - perc)

    return vmin, vmax


def plot_training_scores(adata_map, bins=10, alpha=0.7):
    """
    Plots the 4-panel training diagnosis plot
    Args:
        adata_map (AnnData):
        bins (int or string): Optional. Default is 10.
        alpha (float): Optional. Ranges from 0-1, and controls the opacity. Default is 0.7.
    Returns:
        None
    """
    fig, axs = plt.subplots(1, 4, figsize=(12, 3), sharey=True)
    df = adata_map.uns["train_genes_df"]
    axs_f = axs.flatten()

    # set limits for axis
    axs_f[0].set_ylim([0.0, 1.0])
    for i in range(1, len(axs_f)):
        axs_f[i].set_xlim([0.0, 1.0])
        axs_f[i].set_ylim([0.0, 1.0])

    #     axs_f[0].set_title('Training scores for single genes')
    sns.histplot(data=df, y="train_score", bins=bins, ax=axs_f[0], color="coral")

    axs_f[1].set_title("score vs sparsity (single cells)")
    sns.scatterplot(
        data=df,
        y="train_score",
        x="sparsity_sc",
        ax=axs_f[1],
        alpha=alpha,
        color="coral",
    )

    axs_f[2].set_title("score vs sparsity (spatial)")
    sns.scatterplot(
        data=df,
        y="train_score",
        x="sparsity_sp",
        ax=axs_f[2],
        alpha=alpha,
        color="coral",
    )

    axs_f[3].set_title("score vs sparsity (sp - sc)")
    sns.scatterplot(
        data=df,
        y="train_score",
        x="sparsity_diff",
        ax=axs_f[3],
        alpha=alpha,
        color="coral",
    )

    plt.tight_layout()


def plot_gene_sparsity(
    adata_1, adata_2, xlabel="adata_1", ylabel="adata_2", genes=None, s=1
):
    """
    Compare sparsity of all genes between `adata_1` and `adata_2`.
    Args:
        adata_1 (AnnData): Input data
        adata_2 (AnnData): Input data
        xlabel (str): Optional. For setting the xlabel in the plot. Default is 'adata_1'.
        ylabel (str): Optional. For setting the ylabel in the plot. Default is 'adata_2'.  
        genes (list): Optional. List of genes to use. If `None`, all genes are used.
        s (float): Optional. Controls the size of marker. Default is 1.
    Returns:
        None
    """
    logging.info("Pre-processing AnnDatas...")
    adata_1, adata_2 = mu.pp_adatas(adata_1, adata_2, genes=genes)
    assert adata_1.uns["training_genes"] == adata_2.uns["training_genes"]
    training_genes = adata_1.uns["training_genes"]

    logging.info("Annotating sparsity...")
    annotate_gene_sparsity(adata_1)
    annotate_gene_sparsity(adata_2)
    xs = adata_1[:, training_genes].var["sparsity"].values
    ys = adata_2[:, training_genes].var["sparsity"].values
    fig, ax = plt.subplots(1, 1)
    ax.set_aspect(1)
    ax.set_xlabel("sparsity (" + xlabel + ")")
    ax.set_ylabel("sparsity (" + ylabel + ")")
    ax.set_title(f"Gene sparsity ({len(xs)} genes)")
    ax.scatter(xs, ys, s=s, marker="x")


def ordered_predictions(xs, ys, preds, reverse=False):
    """
    Utility function that orders 2d points based on values associated to each point.
    Args:
        xs (Pandas series): Sequence of x coordinates (floats).
        ys (Pandas series): Sequence of y coordinates (floats).
        preds (Pandas series): Sequence of spatial prediction.
        reverse (bool): Optional. False will sort ascending, True will sort descending. Default is False.
        
    Returns:
        Returns the ordered xs, ys, preds.
    """
    assert len(xs) == len(ys) == len(preds)
    return list(
        zip(
            *[
                (x, y, z)
                for x, y, z in sorted(
                    zip(xs, ys, preds), key=lambda pair: pair[2], reverse=reverse
                )
            ]
        )
    )


def convert_adata_array(adata):
    if isinstance(adata.X, csc_matrix) or isinstance(adata.X, csr_matrix):
        adata.X = adata.X.toarray()


def construct_obs_plot(df_plot, adata, perc=0, suffix=None):
    # clip
    df_plot = df_plot.clip(df_plot.quantile(perc), df_plot.quantile(1 - perc), axis=1)

    # normalize
    df_plot = (df_plot - df_plot.min()) / (df_plot.max() - df_plot.min())

    if suffix:
        df_plot = df_plot.add_suffix(" ({})".format(suffix))
    adata.obs = pd.concat([adata.obs, df_plot], axis=1)


def plot_cell_annotation_sc(
    adata_sp, 
    annotation_list, 
    x="x", 
    y="y", 
    spot_size=None, 
    scale_factor=None, 
    perc=0,
    alpha_img=1.0,
    bw=False,
    ax=None
):
        
    # remove previous df_plot in obs
    adata_sp.obs.drop(annotation_list, inplace=True, errors="ignore", axis=1)

    # construct df_plot
    df = adata_sp.obsm["tangram_ct_pred"][annotation_list]
    construct_obs_plot(df, adata_sp, perc=perc)
    
    #non visium data 
    if 'spatial' not in adata_sp.obsm.keys():
        #add spatial coordinates to obsm of spatial data 
        coords = [[x,y] for x,y in zip(adata_sp.obs[x].values,adata_sp.obs[y].values)]
        adata_sp.obsm['spatial'] = np.array(coords)
    
    if 'spatial' not in adata_sp.uns.keys() and spot_size == None and scale_factor == None:
        raise ValueError("Spot Size and Scale Factor cannot be None when ad_sp.uns['spatial'] does not exist")
    
    #REVIEW
    if 'spatial' in adata_sp.uns.keys() and spot_size != None and scale_factor != None:
        raise ValueError("Spot Size and Scale Factor should be None when ad_sp.uns['spatial'] exists")
    
    sc.pl.spatial(
        adata_sp, color=annotation_list, cmap="viridis", show=False, frameon=False, spot_size=spot_size,
        scale_factor=scale_factor, alpha_img=alpha_img, bw=bw, ax=ax
    )

    adata_sp.obs.drop(annotation_list, inplace=True, errors="ignore", axis=1)


def plot_cell_annotation(
    adata_map,
    adata_sp,
    annotation="cell_type",
    x="x",
    y="y",
    nrows=1,
    ncols=1,
    s=5,
    cmap="viridis",
    subtitle_add=False,
    robust=False,
    perc=0,
    invert_y=True,
):
    """
    Transfer an annotation for a single cell dataset onto space, and visualize
    corresponding spatial probability maps.
    Args:
        adata_map (AnnData): cell-by-spot AnnData containing mapping result
        adata_sp (AnnData): spot-by-gene spatial AnnData
        annotation (str): Optional. Must be a column in `adata_map.obs`. Default is 'cell_type'.
        x (str): Optional. Column name for spots x-coordinates (must be in `adata_map.var`). Default is 'x'.
        y (str): Optional. Column name for spots y-coordinates (must be in `adata_map.var`). Default is 'y'.
        nrows (int): Optional. Number of rows of the subplot grid. Default is 1.
        ncols (int): Optional. Number of columns of the subplot grid. Default is 1.
        s (float): Optional. Marker size. Default is 5.
        cmap (str): Optional. Name of colormap. Default is 'viridis'.
        subtitle_add (bool): Optional. If add annotation name as the subtitle. Default is False.
        robust (bool): Optional. If True, the colormap range is computed with given percentiles instead of extreme values.
        perc (float): Optional. percentile used to calculate colormap range, only used when robust is True. Default is zero.
        invert_y (bool): Optional. If invert the y axis for the plot. Default is True.
    Returns:
        None
    """

    # TODO ADD CHECKS for x and y

    if not robust and perc != 0:
        raise ValueError("Arg perc is zero when robust is False.")

    if robust and perc == 0:
        raise ValueError("Arg perc cannot be zero when robust is True.")

    project_cell_annotations(adata_map, adata_sp, annotation=annotation)

    df_annotation = adata_sp.obsm["tangram_ct_pred"]

    #### Colorbar:
    fig, ax = plt.subplots(figsize=(4, 0.4))
    fig.subplots_adjust(top=0.5)

    cmap = plt.get_cmap(cmap)
    norm = mpl.colors.Normalize(vmin=0, vmax=1,)

    cb1 = mpl.colorbar.ColorbarBase(
        ax, cmap=cmap, norm=norm, orientation="horizontal", label="Probability"
    )
    #### Colorbar

    if nrows is None or ncols is None:
        ncols = 1
        nrows = len(df_annotation.columns)

    fig, axs = plt.subplots(
        nrows, ncols, figsize=(ncols * 3, nrows * 3), sharex=True, sharey=True
    )

    axs_f = axs.flatten()
    if invert_y is True:
        axs_f[0].invert_yaxis()

    [ax.axis("off") for ax in axs_f]

    if len(df_annotation.columns) > nrows * ncols:
        logging.warning(
            "Number of panels smaller than annotations. Increase `nrows`/`ncols`."
        )

    iterator = zip(df_annotation.columns, range(nrows * ncols))
    for ann, index in iterator:
        xs, ys, preds = ordered_predictions(
            adata_map.var[x], adata_map.var[y], df_annotation[ann]
        )

        if robust:
            vmin, vmax = q_value(preds, perc=perc)
        else:
            vmin, vmax = q_value(preds, perc=0)

        axs_f[index].scatter(x=xs, y=ys, c=preds, s=s, cmap=cmap, vmin=vmin, vmax=vmax)
        axs_f[index].set_title(ann)
        axs_f[index].set_aspect(1)

    if subtitle_add is True:
        fig.suptitle(annotation)

def plot_cell_annotation_sc1(
    adata_sp, 
    annotation_list, 
    x="x", 
    y="y", 
    spot_size=None, 
    scale_factor=0.1, 
    perc=0,
    ncols=5,
    cmap='RdBu_r',
    ax=None
):
    
    fig = plt.figure()
    # remove previous df_plot in obs
    adata_sp.obs.drop(annotation_list, inplace=True, errors="ignore", axis=1)
    # construct df_plot
    df = adata_sp.obsm["tangram_ct_pred"][annotation_list]
    construct_obs_plot(df, adata_sp, perc=perc)
    sc.pl.spatial(
        adata_sp, color=annotation_list, show=False, frameon=False, spot_size=spot_size, scale_factor=scale_factor, ax=ax, cmap=cmap, ncols=ncols
    )


def plot_genes_sc(
    genes, 
    adata_measured, 
    adata_predicted,
    x="x",
    y = "y",
    spot_size=None, 
    scale_factor=None, 
    cmap="inferno", 
    perc=0,
    alpha_img=1.0,
    bw=False,
    return_figure=False
):

    # remove df_plot in obs
    adata_measured.obs.drop(
        ["{} (measured)".format(gene) for gene in genes],
        inplace=True,
        errors="ignore",
        axis=1,
    )
    adata_predicted.obs.drop(
        ["{} (predicted)".format(gene) for gene in genes],
        inplace=True,
        errors="ignore",
        axis=1,
    )

    # prepare adatas
    convert_adata_array(adata_measured)

    adata_measured.var.index = [g.lower() for g in adata_measured.var.index]
    adata_predicted.var.index = [g.lower() for g in adata_predicted.var.index]

    adata_predicted.obsm = adata_measured.obsm
    adata_predicted.uns = adata_measured.uns

    # remove previous df_plot in obs
    adata_measured.obs.drop(
        ["{} (measured)".format(gene) for gene in genes],
        inplace=True,
        errors="ignore",
        axis=1,
    )
    adata_predicted.obs.drop(
        ["{} (predicted)".format(gene) for gene in genes],
        inplace=True,
        errors="ignore",
        axis=1,
    )

    # construct df_plot
    data = []
    for ix, gene in enumerate(genes):
        if gene not in adata_measured.var.index:
            data.append(np.zeros_like(np.array(adata_measured[:, 0].X).flatten()))
        else:
            data.append(np.array(adata_measured[:, gene].X).flatten())

    df = pd.DataFrame(
        data=np.array(data).T, columns=genes, index=adata_measured.obs.index,
    )
    construct_obs_plot(df, adata_measured, suffix="measured")

    df = pd.DataFrame(
        data=np.array(adata_predicted[:, genes].X),
        columns=genes,
        index=adata_predicted.obs.index,
    )
    construct_obs_plot(df, adata_predicted, perc=perc, suffix="predicted")

    fig = plt.figure(figsize=(7, len(genes) * 3.5))
    gs = GridSpec(len(genes), 2, figure=fig)
    
    #non visium data
    if 'spatial' not in adata_measured.obsm.keys():
        #add spatial coordinates to obsm of spatial data 
        coords = [[x,y] for x,y in zip(adata_measured.obs[x].values,adata_measured.obs[y].values)]
        adata_measured.obsm['spatial'] = np.array(coords)
        coords = [[x,y] for x,y in zip(adata_predicted.obs[x].values,adata_predicted.obs[y].values)]
        adata_predicted.obsm['spatial'] = np.array(coords)

    if ("spatial" not in adata_measured.uns.keys()) and (spot_size==None and scale_factor==None):
        raise ValueError("Spot Size and Scale Factor cannot be None when ad_sp.uns['spatial'] does not exist")
        
    for ix, gene in enumerate(genes):
        ax_m = fig.add_subplot(gs[ix, 0])
        sc.pl.spatial(
            adata_measured,
            spot_size=spot_size,
            scale_factor=scale_factor,
            color=["{} (measured)".format(gene)],
            frameon=False,
            ax=ax_m,
            show=False,
            cmap=cmap,
            alpha_img=alpha_img,
            bw=bw
        )
        ax_p = fig.add_subplot(gs[ix, 1])
        sc.pl.spatial(
            adata_predicted,
            spot_size=spot_size,
            scale_factor=scale_factor,
            color=["{} (predicted)".format(gene)],
            frameon=False,
            ax=ax_p,
            show=False,
            cmap=cmap,
            alpha_img=alpha_img,
            bw=bw
        )
        
    #     sc.pl.spatial(adata_measured, color=['{} (measured)'.format(gene) for gene in genes], frameon=False)
    #     sc.pl.spatial(adata_predicted, color=['{} (predicted)'.format(gene) for gene in genes], frameon=False)

    # remove df_plot in obs
    adata_measured.obs.drop(
        ["{} (measured)".format(gene) for gene in genes],
        inplace=True,
        errors="ignore",
        axis=1,
    )
    adata_predicted.obs.drop(
        ["{} (predicted)".format(gene) for gene in genes],
        inplace=True,
        errors="ignore",
        axis=1,
    )
    if return_figure==True:
        return fig


def plot_genes(
    genes,
    adata_measured,
    adata_predicted,
    x="x",
    y="y",
    s=5,
    log=False,
    cmap="inferno",
    robust=False,
    perc=0,
    invert_y=True,
):
    """
    Utility function to plot and compare original and projected gene spatial pattern ordered by intensity of the gene signal.
    Args:
        genes (list): list of gene names (str).
        adata_measured (AnnData): ground truth gene spatial AnnData
        adata_predicted (AnnData): projected gene spatial AnnData, can also be adata_ge_cv AnnData returned by cross_validation under 'loo' mode
        x (str): Optional. Column name for spots x-coordinates (must be in `adata_measured.var` and `adata_predicted.var`). Default is 'x'.
        y (str): Optional. Column name for spots y-coordinates (must be in `adata_measured.var` and `adata_predicted.var`). Default is 'y'.
        s (float): Optional. Marker size. Default is 5.
        log: Optional. Whether to apply the log before plotting. Default is False.
        cmap (str): Optional. Name of colormap. Default is 'inferno'.
        robust (bool): Optional. If True, the colormap range is computed with given percentiles instead of extreme values.
        perc (float): Optional. percentile used to calculate colormap range, only used when robust is True. Default is zero.
        invert_y (bool): Optional. If invert the y axis for the plot. Default is True.
    Returns:
        None
    """
    # TODO: not very elegant and slow as hell

    if not robust and perc != 0:
        raise ValueError("Arg perc is zero when robust is False.")

    if robust and perc == 0:
        raise ValueError("Arg perc cannot be zero when robust is True.")

    if isinstance(adata_measured.X, csc_matrix) or isinstance(
        adata_measured.X, csr_matrix
    ):
        adata_measured.X = adata_measured.X.toarray()

    adata_measured.var.index = [g.lower() for g in adata_measured.var.index]
    adata_predicted.var.index = [g.lower() for g in adata_predicted.var.index]

    #### Colorbar:
    fig, ax = plt.subplots(figsize=(4, 0.4))
    fig.subplots_adjust(top=0.5)

    cmap = plt.get_cmap(cmap)
    norm = mpl.colors.Normalize(vmin=0, vmax=1,)

    cb1 = mpl.colorbar.ColorbarBase(
        ax, cmap=cmap, norm=norm, orientation="horizontal", label="Expression Level"
    )
    #### Colorbar

    fig, axs = plt.subplots(nrows=len(genes), ncols=2, figsize=(6, len(genes) * 3))

    for ix, gene in enumerate(genes):
        if gene not in adata_measured.var.index:
            vs = np.zeros_like(np.array(adata_measured[:, 0].X).flatten())
        else:
            vs = np.array(adata_measured[:, gene].X).flatten()

        xs, ys, vs = ordered_predictions(
            adata_measured.obs[x], adata_measured.obs[y], vs,
        )

        if log:
            vs = np.log(1 + np.asarray(vs))
        axs[ix, 0].scatter(xs, ys, c=vs, cmap=cmap, s=s)
        axs[ix, 0].set_title(gene + " (measured)")
        axs[ix, 0].axis("off")
        axs[ix, 0].set_aspect(1)

        xs, ys, vs = ordered_predictions(
            adata_predicted.obs[x],
            adata_predicted.obs[y],
            np.array(adata_predicted[:, gene].X).flatten(),
        )

        if robust:
            vmin, vmax = q_value(vs, perc=perc)
        else:
            vmin, vmax = q_value(vs, perc=0)

        if log:
            vs = np.log(1 + np.asarray(vs))
        axs[ix, 1].scatter(xs, ys, c=vs, cmap=cmap, s=s, vmin=vmin, vmax=vmax)
        axs[ix, 1].set_title(gene + " (predicted)")
        axs[ix, 1].axis("off")
        axs[ix, 1].set_aspect(1)

        if invert_y is True:
            axs[ix, 0].invert_yaxis()
            axs[ix, 1].invert_yaxis()


def quick_plot_gene(
    gene, adata, x="x", y="y", s=50, log=False, cmap="viridis", robust=False, perc=0
):
    """
    Utility function to quickly plot a gene in a AnnData structure ordered by intensity of the gene signal.
    
    Args:
        gene (str): Gene name.
        adata (AnnData): spot-by-gene spatial data.
        x (str): Optional. Column name for spots x-coordinates (must be in `adata.var`). Default is 'x'.
        y (str): Optional. Column name for spots y-coordinates (must be in `adata.var`). Default is 'y'.
        s (float): Optional. Marker size. Default is 5.
        log: Optional. Whether to apply the log before plotting. Default is False.
        cmap (str): Optional. Name of colormap. Default is 'viridis'.
        robust (bool): Optional. If True, the colormap range is computed with given percentiles instead of extreme values.
        perc (float): Optional. percentile used to calculate colormap range, only used when robust is True. Default is zero.
    Returns:
        None
    """
    if not robust and perc != 0:
        raise ValueError("Arg perc is zero when robust is False.")

    if robust and perc == 0:
        raise ValueError("Arg perc cannot be zero when robust is True.")

    xs, ys, vs = ordered_predictions(
        adata.obs[x], adata.obs[y], np.array(adata[:, gene].X).flatten()
    )
    if robust:
        vmin, vmax = q_value(vs, perc=perc)
    else:
        vmin, vmax = q_value(vs, perc=0)
    if log:
        vs = np.log(1 + np.asarray(vs))
    plt.scatter(xs, ys, c=vs, cmap=cmap, s=s, vmin=vmin, vmax=vmax)


def plot_annotation_entropy(adata_map, annotation="cell_type"):
    """
    Utility function to plot entropy box plot by each annotation.
    Args:
        adata_map (AnnData): cell-by-voxel tangram mapping result.
        annotation (str): Optional. Must be a column in `adata_map.obs`. Default is 'cell_type'.
    Returns:
        None
    """
    qk = np.ones(shape=(adata_map.n_obs, adata_map.n_vars))
    adata_map.obs["entropy"] = entropy(adata_map.X, base=adata_map.X.shape[1], axis=1)
    fig, ax = plt.subplots(1, 1, figsize=(10, 3))
    ax.set_ylim(0, 1)
    sns.boxenplot(x=annotation, y="entropy", data=adata_map.obs, ax=ax)
    plt.xticks(rotation=30)


def plot_test_scores(df_gene_score, bins=10, alpha=0.7):
    """
    Plots gene level test scores with each gene's sparsity for mapping result.
    
    Args:
        df_gene_score (Pandas dataframe): returned by compare_spatial_geneexp(adata_ge, adata_sp, adata_sc); 
                       with "gene names" as the index and "score", "sparsity_sc", "sparsity_sp", "sparsity_diff" as the columns
        bins (int or string): Optional. Default is 10.
        alpha (float): Optional. Ranges from 0-1, and controls the opacity. Default is 0.7.
    Returns:
        None
    """

    # check if df_gene_score has all required columns
    if not set(["score", "sparsity_sc", "sparsity_sp", "sparsity_diff"]).issubset(
        set(df_gene_score.columns)
    ):
        raise ValueError(
            "There are missing columns in df_gene_score. Run `compare_spatial_geneexp` with `adata_ge`, `adata_sp`, and `adata_sc` to produce complete dataframe input."
        )

    if "is_training" in df_gene_score.keys():
        df = df_gene_score[df_gene_score["is_training"] == False]
    else:
        df = df_gene_score

    df.rename({"score": "test_score"}, axis="columns", inplace=True)

    fig, axs = plt.subplots(1, 4, figsize=(12, 3), sharey=True)
    axs_f = axs.flatten()

    # set limits for axis
    axs_f[0].set_ylim([0.0, 1.0])
    for i in range(1, len(axs_f)):
        axs_f[i].set_xlim([0.0, 1.0])
        axs_f[i].set_ylim([0.0, 1.0])

    sns.histplot(data=df, y="test_score", bins=bins, ax=axs_f[0])

    axs_f[1].set_title("score vs sparsity (single cells)")
    sns.scatterplot(data=df, y="test_score", x="sparsity_sc", ax=axs_f[1], alpha=alpha)

    axs_f[2].set_title("score vs sparsity (spatial)")
    sns.scatterplot(data=df, y="test_score", x="sparsity_sp", ax=axs_f[2], alpha=alpha)

    axs_f[3].set_title("score vs sparsity (sp - sc)")
    sns.scatterplot(
        data=df, y="test_score", x="sparsity_diff", ax=axs_f[3], alpha=alpha
    )
    plt.tight_layout()

    
def plot_auc(df_all_genes, test_genes=None):
    """
        Plots auc curve which is used to evaluate model performance.
    
    Args:
        df_all_genes (Pandas dataframe): returned by compare_spatial_geneexp(adata_ge, adata_sp); 
        test_genes (list): list of test genes, if not given, test_genes will be set to genes where 'is_training' field is False
    Returns:
        None
    """
    metric_dict, ((pol_xs, pol_ys), (xs, ys)) = eval_metric(df_all_genes, test_genes)
    
    fig = plt.figure()
    plt.figure(figsize=(6, 5))

    plt.plot(pol_xs, pol_ys, c='r')
    sns.scatterplot(xs, ys, alpha=0.5, edgecolors='face')
        
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.gca().set_aspect(.5)
    plt.xlabel('score')
    plt.ylabel('spatial sparsity')
    plt.tick_params(axis='both', labelsize=8)
    plt.title('Prediction on test transcriptome')
    
    textstr = 'auc_score={}'.format(np.round(metric_dict['auc_score'], 3))
    props = dict(boxstyle='round', facecolor='wheat', alpha=0.3)
    # place a text box in upper left in axes coords
    plt.text(0.03, 0.1, textstr, fontsize=11, verticalalignment='top', bbox=props);

    
# Colors used in the manuscript for deterministic assignment.
mapping_colors = {
    "L6 CT": (0.19215686274509805, 0.5098039215686274, 0.7411764705882353),
    "L6 IT": (0.4196078431372549, 0.6823529411764706, 0.8392156862745098),
    "L5/6 NP": (0.6196078431372549, 0.792156862745098, 0.8823529411764706),
    "L6b": "#0000c2ff",
    "L2/3 IT": (0.9019607843137255, 0.3333333333333333, 0.050980392156862744),
    "L5 IT": (0.19215686274509805, 0.6392156862745098, 0.32941176470588235),
    "L5 ET": (0.4549019607843137, 0.7686274509803922, 0.4627450980392157),
    "Oligo": (0.4588235294117647, 0.4196078431372549, 0.6941176470588235),
    "Vip": (0.6196078431372549, 0.6039215686274509, 0.7843137254901961),
    "Astro": "#ffdd55ff",
    "Micro-PVM": "#000000ff",
    "Pvalb": (0.38823529411764707, 0.38823529411764707, 0.38823529411764707),
    "Lamp5": (0.5882352941176471, 0.5882352941176471, 0.5882352941176471),
    "Sst": (0.7411764705882353, 0.7411764705882353, 0.7411764705882353),
    "Sst Chodl": (0.8509803921568627, 0.8509803921568627, 0.8509803921568627),
    "Sncg": (0.5176470588235295, 0.23529411764705882, 0.2235294117647059),
    "Peri": (0.6784313725490196, 0.28627450980392155, 0.2901960784313726),
    "VLMC": (0.8392156862745098, 0.3803921568627451, 0.4196078431372549),
    "Endo": (0.9058823529411765, 0.5882352941176471, 0.611764705882353),
    "Meis2": "#FFA500ff",
    "SMC": "#000000ff",
    "L6 PT": "#4682B4ff",
    "L5 PT": "#a1ed7bff",
    "L5 NP": "#6B8E23ff",
    "L4": "#d61f1dff",
    "Macrophage": "#2b2d2fff",
    "CR": "#000000ff",
}


import json
import os
import pickle
import shutil
import time

import numpy as np
import torch
import torch.nn
import torch.optim as optim

from .ccc_model.ccc_model import CCC
from .ccc_model.process_walks_gcn import (
    Processing_GCN_Walks,
    get_normalized_masked_ids,
)
from .utils.data_utils import (
    create_batches,
    split_data,
)
from .utils.utils import EarlyStopping, load_pickle, show_progress


def setup_pretraining_input(args, attr_graph, context_gen, data_path):
    # split train/validation/test dataset
    tasks = ["train", "validate", "test"]
    split_file = {
        task: os.path.join(data_path, args.data_name + "_pretrain_" + task + ".txt")
        for task in tasks
    }
    if os.path.exists(split_file["train"]):
        print("\n load existing /validate/test data ...")
        for task in tasks:
            fin = open(split_file[task], "r")
            walk_data = []
            for line in fin:
                line = json.loads(line)
                walk_data.append(line)
            fin.close()
            if task == "train":
                walk_train = walk_data
            elif task == "validate":
                walk_validate = walk_data
            else:
                walk_test = walk_data
    else:
        # generate walks
        #print("\n Generating subgraphs for pre-training ...")
        all_walks = context_gen.get_pretrain_subgraphs(
            data_path,
            args.data_name,
            args.beam_width,
            args.max_length,
            args.walk_type,
        )
        #print("\n split data to train/validate/test and save the files ...")
        walk_train, walk_validate, walk_test = split_data(
            all_walks, data_path, args.data_name
        )
    #print(len(walk_train), len(walk_validate), len(walk_test))

    # create batches
    num_batches = {}
    for task in tasks:
        if task == "train":
            walk_data = walk_train
        elif task == "validate":
            walk_data = walk_validate
        else:
            walk_data = walk_test
        cnt = create_batches(
            walk_data, data_path, args.data_name, task, args.batch_size
        )
        num_batches[task] = cnt
    #print("number of batches for pre-training: ", num_batches)
    return num_batches


def run_pretraining(
    args, attr_graph, no_batches, pretrained_node_embedding, ent2id, rel2id
):
    earlystopping = EarlyStopping(patience=3, delta=0.001)
    # data_path = args.data_path +'compgcn_output/' + args.data_name + '/'
    data_path = args.data_path +"/" + args.data_name + "/outdir/"
    out_dir = args.outdir + "/"
    try:
        shutil.rmtree(out_dir)
        os.mkdir(out_dir)
    except:
        os.mkdir(out_dir)

    relations = attr_graph.relation_to_id
    #no_relations = attr_graph.get_number_of_relations()
    nodeid2rowid = attr_graph.get_nodeid2rowid()
    #print(attr_graph.get_number_of_nodes())
    #no_nodes = attr_graph.get_number_of_nodes()
    walk_processor = Processing_GCN_Walks(
        nodeid2rowid, relations, args.n_pred, args.max_length, args.max_pred
    )

    #print("\n processing walks in minibaches before running model:")
    batch_input_file = os.path.join(data_path, args.data_name + "_batch_input.pickled")
    if os.path.exists(batch_input_file):
        #print("loading saved files ...")
        batch_input = load_pickle(batch_input_file)
    else:
        batch_input = {}
        tasks = ["train", "validate", "test"]
        for task in tasks:
            #print(task)
            batch_input[task] = {}
            for batch_id in range(no_batches[task]):
                (
                    subgraphs_list,
                    all_nodes,
                    masked_nodes,
                    masked_postion,
                ) = walk_processor.process_minibatch(
                    data_path, args.data_name, task, batch_id
                )

                batch_input[task][batch_id] = [
                    subgraphs_list,
                    all_nodes,
                    masked_nodes,
                    masked_postion,
                ]
        pickle.dump(batch_input, open(batch_input_file, "wb"))

    if args.is_pre_trained:
        pretrained_node_embedding_tensor = pretrained_node_embedding
    else:
        pretrained_node_embedding_tensor = None

    ccc = CCC(
        args.n_layers,
        args.device,
        args.d_model,
        args.d_k,
        args.d_v,
        args.d_ff,
        args.n_heads,
        attr_graph,
        pretrained_node_embedding_tensor,
        args.is_pre_trained,
        args.base_embedding_dim,
        args.max_length,
        args.num_gcn_layers,
        args.node_edge_composition_func,
        args.gcn_option,
        args.get_bert_encoder_embeddings,
        ent2id,
        rel2id,
    )

    # if torch.cuda.is_available():
    #    ccc =  torch.nn.DataParallel(ccc, device_ids=[0,1])

    criterion = torch.nn.CrossEntropyLoss()
    optimizer = optim.Adam(ccc.parameters(), args.lr)
    node_embeddings = dict()

    loss_dev_final = []
    print("\n Begin Training")
    for epoch in range(args.n_epochs):
        loss_arr = []
        loss_dev_min = 1e6

        start_time = time.time()
        for batch_id in range(no_batches["train"]):
            (
                subgraphs_list,
                all_nodes,
                masked_nodes,
                masked_postion,
            ) = walk_processor.process_minibatch(
                data_path, args.data_name, "train", batch_id
            )

            if args.get_bert_encoder_embeddings:
                logits_lm, _ = ccc(
                    subgraphs_list, all_nodes, masked_postion, masked_nodes
                )
                # node_embeddings[subgraphs_list] = bert_subgraph_embedding_output
            else:
                logits_lm = ccc(
                    subgraphs_list, all_nodes, masked_postion, masked_nodes
                )
            # print('check ====', logits_lm.size(), masked_nodes_id.size())
            try:
                normalized_masked_nodes_id = get_normalized_masked_ids(
                    masked_nodes, ent2id
                )

            except:
                normalized_masked_nodes_id = masked_nodes
            # print(logits_lm.size(), normalized_masked_nodes_id.size())
            loss = criterion(
                logits_lm.transpose(1, 2), normalized_masked_nodes_id.to(args.device)
            )

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            loss_arr.append(loss.data.cpu().numpy().tolist())

            ccnt = batch_id % args.checkpoint
            if batch_id > 0 and batch_id % args.checkpoint == 0:
                ccnt = args.checkpoint
            '''
            message = "Loss: {}, AvgLoss: {}{}".format(
                np.around(loss.data.cpu().numpy().tolist(), 4),
                np.around(np.average(loss_arr).tolist(), 4),
                " " * 10,
            )
            show_progress(ccnt, min(args.checkpoint, no_batches["train"]), message)
            '''
            if (
                batch_id % args.checkpoint == 0 and batch_id > 0
            ) or batch_id == no_batches["train"] - 1:
                '''
                print(
                    "\nBatch: {}, Loss: {}, AvgLoss: {}".format(
                        batch_id,
                        np.around(loss.data.cpu().numpy().tolist(), 4),
                        np.around(np.average(loss_arr).tolist(), 4),
                    )
                )
                '''
                # validation
                ccc.eval()

                with torch.no_grad():
                    loss_dev_arr = []
                    for batch_dev_id in range(no_batches["validate"]):
                        (
                            subgraphs_list,
                            all_nodes,
                            masked_nodes,
                            masked_postion,
                        ) = batch_input["validate"][batch_dev_id]

                        if args.get_bert_encoder_embeddings:
                            logits_lm, _ = ccc(
                                subgraphs_list, all_nodes, masked_postion, masked_nodes
                            )
                            # node_embeddings[subgraphs_list] = bert_subgraph_embedding_output
                        else:
                            logits_lm = ccc(
                                subgraphs_list, all_nodes, masked_postion, masked_nodes
                            )

                        try:
                            normalized_masked_nodes_id = get_normalized_masked_ids(
                                masked_nodes, ent2id
                            )
                        except:
                            normalized_masked_nodes_id = masked_nodes
                        # print(normalized_masked_nodes_id.size())
                        loss = criterion(
                            logits_lm.transpose(1, 2), normalized_masked_nodes_id.to(args.device)
                        )

                        loss_dev_arr.append(loss.data.cpu().numpy().tolist())
                    loss_dev_avg = np.average(loss_dev_arr)
                if epoch%10==0:
                    print("\nEpoch: {}".format(epoch))
                    print(
                        "MinLoss: {}, CurLoss: {}".format(
                            np.around(loss_dev_min, 4), np.around(loss_dev_avg, 4)
                        )
                    )
                if loss_dev_avg < loss_dev_min:
                    loss_dev_min = loss_dev_avg
                    fmodel = open(
                        os.path.join(out_dir, "bert_" + str(epoch) + ".model"), "wb"
                    )
                    torch.save(ccc.state_dict(), fmodel)
                    fmodel.close()

                ccc.train()
                # if(earlystopping.check_early_stopping(loss_dev_avg) == True):
                #    print("Loss not decreasing over 3 epochs exiting")
                #    epoch = args.n_epochs + 1

        loss_dev_final.append(loss_dev_min)
        #print("MinLoss: ", np.around(loss_dev_min, 4))
        end_time = time.time()
        if epoch%10==0:
            print("epoch time: (s)", (end_time - start_time))

    best_epoch = np.argsort(loss_dev_final)[0]
    print("\nBest Epoch: {}".format(best_epoch))
    fbest = open(os.path.join(out_dir, "best_epoch_id.txt"), "w")
    fbest.write(str(best_epoch) + "\n")
    fbest.close()
    np.save(f"{args.outdir}/pretraining_loss.npy", loss_dev_final)

    #print("Begin Testing")
    ccc.eval()
    fl_ = os.path.join(out_dir, "bert_{}.model".format(best_epoch))
    ccc.load_state_dict(
        torch.load(fl_, map_location=lambda storage, loc: storage)
    )

    pred_data = []
    true_data = []
    with torch.no_grad():
        for _ in range(no_batches["test"]):
            subgraphs_list, all_nodes, masked_nodes, masked_postion = batch_input[
                "test"
            ][batch_dev_id]

            if args.get_bert_encoder_embeddings:
                logits, bert_subgraph_embedding_output = ccc(
                    subgraphs_list, all_nodes, masked_postion, masked_nodes
                )
                for i, _ in enumerate(subgraphs_list):
                    str_subgraph = get_str_subgraph(subgraphs_list[i])
                    node_embeddings[str_subgraph] = bert_subgraph_embedding_output[i]
            else:
                logits = ccc(
                    subgraphs_list, all_nodes, masked_postion, masked_nodes
                )
            prob = torch.softmax(logits, dim=2)
            score, index = prob.topk(1, dim=2)

            pred_tmp = index.squeeze(2).data.cpu().numpy().tolist()
            try:
                normalized_masked_nodes_id = get_normalized_masked_ids(
                    masked_nodes, ent2id
                )
            except:
                normalized_masked_nodes_id = masked_nodes
            true_tmp = normalized_masked_nodes_id.data.cpu().numpy().tolist()
            pred_data += pred_tmp
            true_data += true_tmp

    # accu, mse = run_evaluation(pred_data, true_data)
    return pred_data, true_data

    # #Dump node vectors
    # if args.get_node_embedding == False:
    #     print('\n Begin dumping node embeddings ...')
    #     node_embedding = ccc.GetNodeVec()
    #     node_embedding_file = os.path.join(data_path, args.data_name+'_node_embedding.pickled')
    #     pickle.dump(node_embedding, open(node_embedding_file, "wb" ))

    # if(args.get_bert_encoder_embeddings):
    #     torch.save(node_embeddings, "node_embeddings_encoder.txt")


def get_str_subgraph(subgraph):
    str_subgraph = ""
    for edge in subgraph:
        str_subgraph += str(edge[0]) + "_" + str(edge[1]) + "_" + str(edge[2]) + ";"
    return str_subgraph



import json
import pickle
import sys


def load_pickle(filename):
    try:
        with open(str(filename), "rb") as f:
            obj = pickle.load(f)

    except EOFError:
        obj = None

    return obj


def show_progress(curr_, total_, message=""):
    """Display progress."""
    prog_ = int(round(100.0 * float(curr_) / float(total_)))
    dstr = "[" + ">" * int(round(prog_ / 4)) + " " * (25 - int(round(prog_ / 4))) + "]"
    sys.stdout.write("{}{}% {}\r".format(dstr, prog_, message))
    sys.stdout.flush()


class EarlyStopping:
    def __init__(self, patience, delta):
        self.patience = patience
        self.delta = delta
        self.counter = 0
        self.lowest_loss = None

    def check_early_stopping(self, val_loss):
        if self.lowest_loss is None:
            self.lowest_loss = val_loss
            return False

        if val_loss + self.delta <= self.lowest_loss:
            self.counter += 1
            if self.counter >= self.patience:
                # stop model
                return True
        return False


def get_id_map(path):
    with open(path, "r") as f:
        data = json.load(f)
    return data



import random
from collections import defaultdict

import networkx as nx


class NegativeSampleGenerator:
    def __init__(self):
        self.node_role_cluster_map = defaultdict(set)
        self.role_clusters = []
        self.node_type_map = dict()
        self.type_node_map = defaultdict(list)
        self.num_node_types = 0
        self.all_relations = set()
        self.schema_graph = nx.MultiGraph()
        return

    def __update_role_clusters(self, role_cluster):

        overlapping_cluster_indices = []
        for i in range(len(self.role_clusters)):
            overlap = self.role_clusters[i].intersection(role_cluster)
            if len(overlap) > 0:
                overlapping_cluster_indices.append(i)

        if len(overlapping_cluster_indices) == 0:
            self.role_clusters.append(role_cluster)
            return
        else:
            # Merge all overlapping clusters
            super_cluster = set()
            for idx in overlapping_cluster_indices:
                super_cluster.update(self.role_clusters[idx])
            new_role_clusters = [super_cluster]
            for i in range(len(self.role_clusters)):
                if i not in overlapping_cluster_indices:
                    new_role_clusters.append(self.role_clusters[i])
            self.role_clusters = new_role_clusters

        return

    def build_graph_schema(self, graph):

        # graph = self.load_graph(path)

        print("Building node cluster map ...")
        print(
            "Collect roles (relation.src/relation.dst) for each node [O(num_edges)]..."
        )
        for e in graph.edges(data="label"):
            self.node_role_cluster_map[e[0]].add("%s.src" % e[2])
            self.node_role_cluster_map[e[1]].add("%s.dst" % e[2])

        print("Merging role clusters [O(num_nodes*num_node_types)]...")
        for role_cluster in self.node_role_cluster_map.values():
            self.__update_role_clusters(role_cluster)

        print("Assigning node types [O(num_nodes)]...")
        self.num_node_types = len(self.role_clusters)
        for node_id, role_cluster in self.node_role_cluster_map.items():
            overlapping_clusters = []
            for i in range(self.num_node_types):
                if len(self.role_clusters[i].intersection(role_cluster)) > 0:
                    overlapping_clusters.append(i)
            # assert(len(overlapping_clusters) == 1)
            type_id = str(overlapping_clusters[0])
            self.node_type_map[node_id] = type_id
            self.type_node_map[type_id].append(node_id)

        print("Building schema graph [O(E)] ...")
        print("Collecting unique schema edges by storing in set ...")
        schema_edges = set()
        for e in graph.edges(data="label"):
            src_node_type = self.node_type_map[e[0]]
            dst_node_type = self.node_type_map[e[1]]
            relation = e[2]
            schema_edge_key = "%s:%s:%s" % (src_node_type, dst_node_type, relation)
            schema_edges.add(schema_edge_key)

        print("Adding edges in MultiGraph schema_graph")
        for e in schema_edges:
            print("Schema edge : ", e)
            tokens = e.split(":")
            self.schema_graph.add_edge(tokens[0], tokens[1], label=tokens[2])
            self.all_relations.add(tokens[2])

        # return graph

    def get_type_def(self, node_t):
        return self.role_clusters[int(node_t)]

    def get_invalid_neighbor_types(self, node_type_id):
        neighbors = list(self.schema_graph.neighbors(node_type_id))
        all_node_types = set([str(i) for i in range(self.num_node_types)])
        return all_node_types - set([node_type_id] + neighbors)

    def get_invalid_relation_types(self, node_t):
        node_relations = []
        for nbr_t in self.schema_graph.neighbors(node_t):
            multi_edge_data = self.schema_graph.get_edge_data(node_t, nbr_t)
            for i, _ in enumerate(multi_edge_data):
                r = multi_edge_data[i]["label"]
                node_relations.append(r)
        negative_relations = self.all_relations - set(node_relations)
        return negative_relations

    def gen_negative_triples(self, graph, count):

        self.build_graph_schema(graph)
        print("Generating negative triples ...")
        negative_relations = set()
        num_rels = 0
        skip_dup = 0
        print("Generating %d", count)
        for node_id in graph.nodes():

            # print('Iterating over node: %s' % node_id)
            node_t = self.node_type_map[node_id]
            # print('Node type:', self.get_type_def(node_t))
            invalid_node_types = self.get_invalid_neighbor_types(node_t)
            invalid_relation_types = self.get_invalid_relation_types(node_t)

            for invalid_node_t in invalid_node_types:

                # print('Invalid nbr type:', self.get_type_def(invalid_node_t))
                invalid_nbr = random.choice(self.type_node_map[invalid_node_t])
                # print('Invalid nbr: ', invalid_nbr)

                for invalid_relation_t in invalid_relation_types:

                    # print('Invalid relation: %s' % invalid_relation_t)
                    triple = (invalid_relation_t, node_id, invalid_nbr, "0")
                    triple_key = ":".join(triple)

                    if triple_key not in negative_relations:
                        negative_relations.add(triple_key)
                        num_rels += 1
                        if num_rels == count:
                            return [
                                triple_key.split(":")
                                for triple_key in negative_relations
                            ]
                    else:
                        skip_dup += 1

        print(
            "Number of -ve relations: %d skipped duplicates: %d"
            % (len(negative_relations), skip_dup)
        )
        return [triple_key.split(":") for triple_key in negative_relations]

    def load_graph(self, path):
        graph = nx.DiGraph()
        with open(path) as f_in:
            for line in f_in:
                tokens = line.strip().split(" ")
                relation = tokens[0]
                src = tokens[1]
                dst = tokens[2]
                graph.add_edge(src, dst, label=relation)
        return graph



import argparse
import json
import random
import os
import time
import shutil
from typing import List
import warnings
#warnings.filterwarnings('ignore')
import torch
import torch.nn as nn
import numpy as np
from gensim.models import Word2Vec
import shutil
import pandas as pd
import math
import re
import networkx as nx
from .src.processing.attributed_graph import AttributedGraph
import pickle
from .src.ccc_model.process_walks_gcn import Processing_GCN_Walks
from .src.finetuning import (
    run_finetuning_wkfl2,
    run_finetuning_wkfl3,
    setup_finetuning_input,
)
import torch.optim as optim
from torch.autograd import Variable
from .src.pretraining import run_pretraining, setup_pretraining_input
from .src.processing.context_generator import ContextGenerator
from .src.processing.generic_attributed_graph import GenericGraph
from .src.utils.data_utils import load_pretrained_node2vec
from .src.utils.evaluation import run_evaluation_main
from .src.utils.link_predict import find_optimal_cutoff, link_prediction_eval
from .src.utils.utils import get_id_map, load_pickle
from .src.ccc_model.ccc_model import CCC, FinetuneLayer
from .src.ccc_model.process_walks_gcn import Processing_GCN_Walks
from .src.utils.context_metrics import PathMetrics
from .src.utils.data_utils import create_finetune_batches2
from .src.utils.utils import load_pickle, show_progress


def get_set_embeddings_details(args):
    if not args.pretrained_embeddings:
        if args.pretrained_method == "compgcn":
            args.pretrained_embeddings = (
                f"{args.emb_dir}/{args.data_name}/"
                f"act_{args.data_name}_{args.node_edge_composition_func}_500.out"
            )
        elif args.pretrained_method == "node2vec":
            args.base_embedding_dim = 128
            args.pretrained_embeddings = (
                f"{args.emb_dir}/{args.data_name}/{args.data_name}.emd"
            )
    else:
        if args.pretrained_method == "compgcn":
            args.base_embedding_dim = 200
        elif args.pretrained_method == "node2vec":
            args.base_embedding_dim = 128
    return args.pretrained_embeddings, args.base_embedding_dim


def get_graph(args, data_path, false_edge_gen):
    #print("\n Loading graph...")
    attr_graph = GenericGraph(data_path, false_edge_gen)
    context_gen = ContextGenerator(attr_graph, args.num_walks_per_node)

    return attr_graph, context_gen


def get_test_edges(paths: List[str], sep: str):
    # edges = set()
    edges = []
    for path in paths:
        with open(path, "r") as f:
            for line in f:
                tokens = line.strip().split(sep)
                etype = tokens[0]
                source = tokens[1]
                destination = tokens[2]
                label = tokens[3]
                edge = (etype, source, destination, label)
                # edges.add(edge)
                edges.append(edge)

    return edges


class GenericGraph_pre(AttributedGraph):
    def __init__(
        self, main_dir, false_edge_gen, attributes_file="", sample_training_set=False
    ):
        """
        Assumes derived classes will create a networkx graph object along with
        following
        1) self.unique_relations = set()
        2) self.node_attr_dfs = dict()
        3) self.node_types = dict()
        4) self.G  = nx.Graph()
        """
        super().__init__()
        self.main_dir = main_dir
        self.false_edge_gen = false_edge_gen
        # self.load_graph(main_dir + "/kg_final.txt")
        self.load_graph(main_dir + "/test.txt")
        self.create_normalized_node_id_map()

        # get product attributes
        self.node_attr = dict()
        self.node_attr_list = []

        if attributes_file == "":
            for node_id in self.G.nodes():
                self.node_attr[node_id] = dict()
                self.node_attr[node_id]["defattr"] = "True"
            self.node_attr_list = ["defattr"]
        else:
            raise NotImplementedError

        self.node_attr_dfs = {
            "deftype": pd.DataFrame.from_dict(self.node_attr, orient="index")
        }
        self.sample_training_set = sample_training_set
        self.get_nodeid2rowid()
        self.get_rowid2nodeid()
        self.get_rowid2vocabid_map()
        self.set_relation_id()

    def generate_link_prediction_dataset(self, outdir, fr_test_edges):
        #print("In GenericGraph link Prediction Generation")
        false_edge_gen = self.false_edge_gen

        self.test_edges = self.load_test_edges(self.main_dir + "/test.txt")
        return self.test_edges

    def load_graph(self, filename):
        """
        This method should be implemented by all DataLoader classes to fill in
        values for
        1) self.G
        2) self.node_attr_dfs
        3) self.unique_relations
        4) self.node_types
        """
        with open(filename, "r") as f:
            for line in f:
                arr = line.strip().split(" ")
                if len(arr) >= 3:
                    edge_type = arr[0]
                    source = arr[1]
                    target = arr[2]
                    if len(arr) == 3 or (len(arr) == 4 and arr[3] == "1"):
                        self.G.add_edge(source, target, label=edge_type)
                        self.unique_relations.add(edge_type)
                        self.node_types[source] = "deftype"
                        self.node_types[target] = "deftype"
        return

    def load_test_edges(self, filename):
        true_and_false_edges = []
        with open(filename, "r") as f:
            for line in f:
                arr = line.strip().split()
                relation = arr[0]
                source = arr[1]
                target = arr[2]
                if len(arr) == 4:
                    true_and_false_edges.append((relation, source, target, arr[3]))
                else:
                    # GATNE dataset have no false edges for training
                    true_and_false_edges.append((relation, source, target, "1"))
        return true_and_false_edges

    def get_continuous_cols(self):
        return {"deftype": None}

    def get_wide_cols(self):
        return {"deftype": self.node_attr_list}

class ContextGenerator_pre:
    def __init__(
        self,
        attributed_graph_obj,
        num_walks_per_node,
        pretrained_embeddings=None,
        optimal_context_per_relation=None,  # FIXME - removed dangerous empty dict init
    ):
        self.attr_graph = attributed_graph_obj
        self.num_walks_per_node = num_walks_per_node
        self.pretrained_embeddings = pretrained_embeddings
        random.seed(10)

    def get_random_k_nbrs(self, nodeid, exclude_list, k):
        """
        returns randomly sampled 'k' nbrs for a given node as
        list[node_ids]
        if k == -1 :  returns all neighbours
        """
        all_nbrs = list(set(list(nx.all_neighbors(self.attr_graph.G, nodeid))))
        valid_nbrs = all_nbrs
        if exclude_list is None or len(exclude_list) == 0:
            valid_nbrs = all_nbrs
        else:
            valid_nbrs = [x for x in all_nbrs if x not in exclude_list]
        if k >= 0 and len(valid_nbrs) >= k:
            return random.sample(valid_nbrs, k)
        else:
            return all_nbrs

    def gen_random_walk_context(
        self, beam_width, max_num_edges, walk_type, valid_patterns=None
    ):  # FIXME - removed dangerous valid_patterns list init
        """
        Given as input an attributed networkx graph, generates
        random walk based node_context for  each node.
        Returns : list[node_context] == list[list[(source, dest, relation)]]
        Node context is generated by performing random walk around the
        node with beam_width = beam_width and depth = 2
        """
        G = self.attr_graph.get_graph()
        all_contexts = []
        for source in list(G.nodes()):
            # for u,v in nx.bfs_beam_edges(G, source, centrality.get, beam_width):
            # for u,v in nx.bfs_beam_edges(G, source, random_neighbour_goodness, beam_width):
            #    node_context.append((u,v, G.edges[u,v]['label']))
            for _ in range(self.num_walks_per_node):
                if walk_type == "bfs":
                    node_context = self.bfs_beam(source, beam_width)
                elif walk_type == "dfs":
                    node_context = self.random_chain(source, max_num_edges)
                else:
                    print("Unknown context generation strategy, select bfs/dfs")
                all_contexts.append(node_context)
        # valid_patterns = set(['1_1_1', '2_2_2', '1_2_1', '2_1_2'])
        # all_contexts = self.filter_walks_by_patterns(all_contexts, valid_patterns)
        return all_contexts

    def filter_walks_by_patterns(self, walks, valid_patterns):
        # valid_walks = [walk for walk in walks  if self.get_pattern(walk) in
        #               valid_patterns]
        valid_walks = []
        for walk in walks:
            walk_pattern = self.get_pattern(walk)
            if walk_pattern in valid_patterns:
                valid_walks.append(walk)
        return valid_walks

    def get_pattern(self, walk, first_last_same=False):
        """
        given a path = list[(source, target, relation)]
        returns a pattern = [source[0].type, source[1].type, source[2].type...]
        """
        pattern = []
        for edge in walk:
            # source = edge[0]
            relation = edge[2]
            # source_type = self.attr_graph.get_node_type(source)
            # pattern.append(source_type)
            pattern.append(relation)
        return "_".join(pattern)
        # return pattern

    def random_chain(self, source, max_num_edges):
        # generates a random walk of length=max_num_edges, without cycles
        # starting from source
        path = []
        nodes_so_far = []
        G = self.attr_graph.get_graph()
        for _ in range(max_num_edges):
            nbr = self.get_random_k_nbrs(source, exclude_list=nodes_so_far, k=1)[0]
            try:
                relation = G.edges[source, nbr]["label"]
            except KeyError:  # FIXME - replaced bare except
                relation = G.edges[nbr, source]["label"]
            path.append((source, nbr, relation))
            nodes_so_far.append(source)
            source = nbr
        return path

    def bfs_beam(self, source, beam_width):
        """
        Given a source node, a beam width and depth=hops
        return a subgraph around source node .
        The subgraph is retuned as [(source, target, relation)]
        """
        subgraph = list()
        G = self.attr_graph.get_graph()
        source_nbrs = self.get_random_k_nbrs(source, exclude_list=[], k=beam_width)
        for x in source_nbrs:
            try:
                relation = G.edges[source, x]["label"]
            except KeyError:  # FIXME - replaced bare except
                relation = G.edges[x, source]["label"]
            subgraph.append((source, x, relation))
        for src_nbr in source_nbrs:
            src_nbr_hop2_cands = self.get_random_k_nbrs(
                src_nbr, exclude_list=[source], k=beam_width
            )
            for x in src_nbr_hop2_cands:
                try:
                    relation = G.edges[src_nbr, x]["label"]
                except KeyError:  # FIXME - replaced bare except
                    relation = G.edges[x, src_nbr]["label"]
                subgraph.append((src_nbr, x, relation))
        # print(len(subgraph))
        return subgraph

    def get_pretrain_subgraphs(
        self,
        data_path,
        data_name,
        beam_width,
        max_num_edges,
        walk_type,
        mode="pretrain",
    ):
        walk_file = os.path.join(data_path, data_name + "_walks.txt")
        if os.path.exists(walk_file):
            #print("\n load walks ...")
            fin = open(walk_file, "r")
            all_walks = []
            for line in fin:
                line = json.loads(line)
                all_walks.append(line)
            fin.close()
        else:
            #print("\n generate walks ...")
            # context_gen = ContextGenerator(attr_graph)
            # all_walks = self.gen_random_walk_context(beam_width)
            # for pretraining should we geerate walks without taking out test
            # and validation
            all_walks = self.gen_random_walk_context(
                beam_width, max_num_edges, walk_type
            )
            walk_file = os.path.join(data_path, data_name + "_walks.txt")
            fout = open(walk_file, "w")
            for walk in all_walks:
                json.dump(walk, fout)
                fout.write("\n")
            fout.close()
        #print("no. of walks", len(all_walks))
        return all_walks

    def get_finetune_subgraphs(
        self, test_edges, beam_width, max_seq_len, option
    ):

        # print("size of train, valid, test set", len(train_edges),
        #      len(valid_edges), len(test_edges))
        #true_test_edges = [x for x in test_edges if x[3] == "1"]
        #true_valid_edges = [x for x in valid_edges if x[3] == "1"]
        #to_be_removed_edges = true_test_edges + true_valid_edges
        #to_be_removed_source_target_pairs = [(x[1], x[2]) for x in to_be_removed_edges]
        #self.attr_graph.G.remove_edges_from(to_be_removed_source_target_pairs)
        walks_by_task_dict = dict()
        for task, task_edges in zip(
            ["test"], [test_edges]
        ):
            task_subgraphs = []
            for edge in task_edges:
                source = edge[1]
                target = edge[2]
                relation = edge[0]
                is_true = edge[3]
                # source_target_contexts = self.get_path_with_edge_label(source, target,
                #                                      max_seq_len-1)
                source_target_contexts = self.get_selective_context(
                    source, target, beam_width, max_seq_len - 1, option
                )
                if len(source_target_contexts) == 0:
                    first_edge = (source, target, relation, is_true)
                    task_subgraphs.append([first_edge])
                else:
                    for subgraph in source_target_contexts:
                        first_edge = (source, target, relation, is_true)
                        subgraph.insert(0, first_edge)
                        task_subgraphs.append(subgraph)
                walks_by_task_dict[task] = task_subgraphs
        # print(walks_by_task_dict)
        # add edges back to graph, in case attributed graph is used by rest of
        # code
        return walks_by_task_dict

    def get_context(self, source, target, beam_width, max_seq_len):
        all_contexts = []
        # for i in range(self.num_walks_per_node):
        source_context = self.bfs_beam(source, beam_width)[0:max_seq_len]
        # target_context = self.bfs_beam(target, beam_width)[0:max_seq_len]
        all_contexts.append(source_context)
        # all_contexts.append(target_context)
        return all_contexts

    def get_selective_context(
        self, source, target, beam_width, max_seq_len, option, valid_patterns=None
    ):  # FIXME - removed dangerous default list init in valid_patterns

        G = self.attr_graph.G
        paths = []
        if option == "shortest":
            try:
                path_node_list = nx.bidirectional_shortest_path(G, source, target)
                path_with_edge_label = []
                for i in range(len(path_node_list) - 1):
                    u = path_node_list[i]
                    v = path_node_list[i + 1]
                    path_with_edge_label.append((u, v, G.edges[u, v]["label"]))
                paths = [path_with_edge_label[0:max_seq_len]]
            except nx.exception.NetworkXNoPath:
                return []

        elif option == "all":
            paths = self.get_path_with_edge_label(source, target, max_seq_len)
            # print("found all paths", source, target, paths)
        elif option == "pattern":
            paths = self.get_path_with_edge_label(source, target, max_seq_len)
            if len(valid_patterns) > 0:
                paths = self.filter_walks_by_patterns(paths, valid_patterns)
            #print("found pattern paths", source, target, paths)
        elif option == "random":
            node_list_paths = self.get_random_paths_between_nodes(
                source, target, beam_width, max_seq_len, currpath=[]
            )
            paths = []
            for node_list_path in node_list_paths:
                path_with_edge_label = []
                node_list_path.insert(0, source)
                for i in range(len(node_list_path) - 1):
                    u = node_list_path[i]
                    v = node_list_path[i + 1]
                    path_with_edge_label.append((u, v, G.edges[u, v]["label"]))
                paths.append(path_with_edge_label)
        elif option == "default":
            paths = self.get_context(source, target, beam_width, max_seq_len)
        return paths

    def get_random_paths_between_nodes(
        self, source, target, beam_width, max_seq_len, currpath
    ):
        """
        Generates random path between source and target , by selecting
        'beam_width' number of neighbors and upto length >= max_seq_len
        """
        if len(currpath) > max_seq_len or source == target:
            return []
        all_paths = []
        exclude_list = []

        # print("In generate paths", source, target, currpath)

        source_nbrs = self.get_random_k_nbrs(source, exclude_list, beam_width)
        # print("selected source nbrs", source_nbrs)
        if target in source_nbrs:
            # print("Found path ending in target", source_nbrs, target)
            path = [target]
            all_paths.append(path)

        for n in source_nbrs:
            if n != target and n not in currpath:
                new_source = n
                new_path = list(currpath)
                new_path.append(new_source)
                nbr_paths = self.get_random_paths_between_nodes(
                    new_source, target, beam_width, max_seq_len, new_path
                )
                new_path = new_path[0:-1]
                for p in nbr_paths:
                    p.insert(0, new_source)
                all_paths.extend(nbr_paths)
        return all_paths

    def get_path_metrics(self, G, source, target, paths):
        """
        Metrics per node pair:
            1) Number of paths connecting the nodes
            2) Number of metapaths
            5) Polysemy behavior : Variation in number of unique relations each node participates in
            6)
        """
        return

    def get_path_with_edge_label(self, source, target, max_seq_len):
        G = self.attr_graph.G
        paths = nx.all_simple_paths(G, source, target, cutoff=max_seq_len)
        paths_with_edge_labels = []
        for path in map(nx.utils.pairwise, paths):
            path_with_edge_label = []
            for u, v in path:
                path_with_edge_label.append((u, v, G.edges[u, v]["label"]))
            paths_with_edge_labels.append(path_with_edge_label)
        return paths_with_edge_labels

def get_graph(args, data_path, false_edge_gen):
    #print("\n Loading graph...")
    attr_graph = GenericGraph_pre(data_path, false_edge_gen)
    context_gen = ContextGenerator_pre(attr_graph, args.num_walks_per_node)

    return attr_graph, context_gen

def split_data_pre(data_list, data_path, data_name):
    random.shuffle(data_list)
    data_size = len(data_list)
    data_test = data_list
    tasks = ["test"]
    split_file = {
        task: os.path.join(data_path, data_name + "_walks_" + task + ".txt")
        for task in tasks
    }
    for task in tasks:
        #print(task, split_file[task])
        fout = open(split_file[task], "w")
        if task == "train":
            walk_data = data_train
        elif task == "validate":
            walk_data = data_validate
        else:
            walk_data = data_test

        for walk in walk_data:
            json.dump(walk, fout)
            fout.write("\n")
        fout.close()

    return data_test

def create_batches_pre(data_list, data_path, data_name, task, batch_size):
    folder = os.path.join(data_path, task)
    if os.path.exists(folder):
        cnt = math.ceil(len(data_list) / batch_size)
        return cnt
    else:
        os.mkdir(folder)
        random.shuffle(data_list)
        data_size = len(data_list)
        cnt = 0
        data = []
        for ii in range(data_size):
            data.append(data_list[ii])
            if len(data) == batch_size:
                file_name = data_name + "_" + task + "_batch_" + str(cnt) + ".txt"
                fout = open(os.path.join(folder, file_name), "w")
                for jj, _ in enumerate(data):
                    json.dump(data[jj], fout)
                    fout.write("\n")
                fout.close()
                data = []
                cnt += 1

        if len(data) > 0:
            file_name = data_name + "_" + task + "_batch_" + str(cnt) + ".txt"
            fout = open(os.path.join(folder, file_name), "w")
            for jj, _ in enumerate(data):
                json.dump(data[jj], fout)
                fout.write("\n")
            fout.close()
            cnt += 1

        return cnt

def setup_pretraining_input_pre(args, attr_graph, context_gen, data_path):
    # split train/validation/test dataset
    tasks = ["test"]
    #print("\n Generating subgraphs for pre-training ...")
    all_walks = context_gen.get_pretrain_subgraphs(
        data_path,
        args.data_name,
        args.beam_width,
        args.max_length,
        args.walk_type,
    )
    #print("\n split data to train/validate/test and save the files ...")
    walk_test = split_data_pre(
        all_walks, data_path, args.data_name
    )
    #print(len(walk_test))

    # create batches
    num_batches = {}
    for task in tasks:
        if task == "train":
            walk_data = walk_train
        elif task == "validate":
            walk_data = walk_validate
        else:
            walk_data = walk_test
        cnt = create_batches_pre(
            walk_data, data_path, args.data_name, task, args.batch_size
        )
        num_batches[task] = cnt
    #print("number of batches for pre-training: ", num_batches)
    return num_batches

def create_finetune_batches2_pre(finetune_data, finetune_path, data_name, batch_size):

    no_batches = {}
    tasks = ["test"]
    #task_file = {task: os.path.join(finetune_path, task + ".txt") for task in tasks}
    for task in tasks:
        #print(task)
        all_task_subgraphs = finetune_data[task]
        tmp_no = create_batches_pre(
            all_task_subgraphs, finetune_path, data_name, task, batch_size
        )
        no_batches[task] = tmp_no
    return no_batches

def setup_finetuning_input_pre(args, attr_graph, context_gen):
    # finetune
    #print("\n Generate data for finetuning ...")
    num_batches = {}
    finetune_path = args.outdir + "/finetune/"
    if not os.path.exists(finetune_path):
        os.mkdir(finetune_path)

    if os.path.exists(finetune_path + "/finetune_walks.txt"):
        finetune_walks_per_task = json.load(open(finetune_path + "/finetune_walks.txt"))
    else:
        test_edges = attr_graph.generate_link_prediction_dataset(
            finetune_path, fr_test_edges=0.1)
        
        finetune_walks_per_task = context_gen.get_finetune_subgraphs(
            test_edges,
            args.beam_width,
            args.max_length,
            args.path_option,
        )
        json.dump(
            finetune_walks_per_task, open(finetune_path + "/finetune_walks.txt", "w")
        )

        #print(
        #    "Number of train, valid, test",
        #    len(test_edges),
        #)
    num_batches = create_finetune_batches2_pre(
        finetune_walks_per_task, finetune_path, args.data_name, args.ft_batch_size
    )
    #print("No. of batches for finetuning:", num_batches)

    return num_batches


def Predict(
    args, attr_graph, no_batches, pretrained_node_embedding, ent2id, rel2id,model_path,best_id,is_trained,model_name
):

    # data_path = args.data_path + args.data_name + "/"
    outdir = args.outdir + "/"
    finetune_path = outdir + "finetune/"
    ft_out_dir = finetune_path + "/results"
    #try:
    #    shutil.rmtree(ft_out_dir)
    #   os.mkdir(ft_out_dir)
    #except:  # FIXME - need to replace bare except
    #   os.mkdir(ft_out_dir)
    relations = attr_graph.relation_to_id
    nodeid2rowid = attr_graph.get_nodeid2rowid()
    walk_processor = Processing_GCN_Walks(
        nodeid2rowid, relations, args.n_pred, args.max_length, args.max_pred
    )

    # process minibatch for finetune(ft)
    #print("\n processing minibaches for finetuning:")
    ft_batch_input_file = os.path.join(
        finetune_path, args.data_name + "_ft_batch_input.pickled"
    )
    if os.path.exists(ft_batch_input_file):
        #print("loading saved files ...")
        ft_batch_input = load_pickle(ft_batch_input_file)
    else:
        ft_batch_input = {}
        tasks = ["test"]
        for task in tasks:
            #print(task)
            ft_batch_input[task] = {}
            for batch_id in range(no_batches[task]):
                (
                    subgraphs,
                    all_nodes,
                    labels,
                ) = walk_processor.process_finetune_minibatch(
                    finetune_path, args.data_name, task, batch_id
                )
                ft_batch_input[task][batch_id] = [subgraphs, all_nodes, labels]
        pickle.dump(ft_batch_input, open(ft_batch_input_file, "wb"))

    if args.is_pre_trained:
        pretrained_node_embedding_tensor = pretrained_node_embedding
    else:
        pretrained_node_embedding_tensor = None

    ccc = CCC(
        args.n_layers,
        args.device,
        args.d_model,
        args.d_k,
        args.d_v,
        args.d_ff,
        args.n_heads,
        attr_graph,
        pretrained_node_embedding_tensor,
        args.is_pre_trained,
        args.base_embedding_dim,
        args.max_length,
        args.num_gcn_layers,
        args.node_edge_composition_func,
        args.gcn_option,
        args.get_bert_encoder_embeddings,
        ent2id,
        rel2id,
    )

    pretrained_node_embedding_tensor = ccc.gcn_graph_encoder.node_embedding

    ft_linear = FinetuneLayer(
        args.device,
        args.d_model,
        args.ft_d_ff,
        args.ft_layer,
        args.ft_drop_rate,
        attr_graph,
        args.ft_input_option,
        args.n_layers,
    )
    

    # run in fine tuning mode
    ccc.set_fine_tuning()

    # testing
    print("Begin Predicting")
    ccc.eval()
    ft_linear.eval()
    if is_trained:
        fl_ = model_name
    else:
        fl_ = os.path.join(model_path, "finetune_{}.model".format(best_id))
    ft_linear.load_state_dict(
        torch.load(fl_, map_location=lambda storage, loc: storage)
    )

    pred_data = []
    true_data = []

    with torch.no_grad():
        att_weights = {}
        final_emd = {}
        for batch_test_id in range(no_batches["test"]):
            #if batch_test_id % 100 == 0:
                #print('###')
                #print("Evaluating test batch {}".format(batch_test_id))
            #print(batch_test_id)
            subgraphs, all_nodes, labels = ft_batch_input["test"][batch_test_id]
            masked_pos = torch.randn(args.batch_size, 2)
            masked_nodes = Variable(
                torch.LongTensor([[] for ii in range(args.ft_batch_size)])
            )
            _, layer_output, att_output = ccc(
                subgraphs, all_nodes, masked_pos, masked_nodes
            )
            score, src_embedding, dst_embedding = ft_linear(layer_output)
            score = score.data.cpu().numpy().tolist()
            labels = labels.data.cpu().numpy().tolist()

            for ii, _ in enumerate(score):
                pred_data.append(score[ii][0])
                true_data.append(labels[ii][0])
            # print(len(score), len(labels))


    return pred_data, true_data



def run_predict(data_name='Apoe_Grm5',data_path='./Test/single-cell/data/',
    outdir='./Test/single-cell/data/Apoe_Grm5/output/',
    pretrained_embeddings='./Test/single-cell/data/data_pca.emd', model_path = './Test/single-cell/data',best_id = 9, is_trained = False,model_name = 'Mif_Epha5.model'):

    #torch.manual_seed(1234)
    parser = argparse.ArgumentParser()
    parser.add_argument("--device", default="cuda:0", type=str, help='specify device')
    parser.add_argument("--data_name", default="EFNA5_EPHA2", help="name of the dataset")
    parser.add_argument("--data_path", default="./data/", help="path to dataset")
    parser.add_argument("--outdir", default="./data/EFNA5_EPHA2/output", help="path to output dir")
    parser.add_argument(
        "--pretrained_embeddings", default="./data_pca.emd",help="absolute path to pretrained embeddings"
    )
    parser.add_argument(
        "--pretrained_method", default="node2vec", help="compgcn|node2vec"
    )
    # Walks options
    parser.add_argument(
        "--beam_width",
        default=1,
        type=int,
        help="beam width used for generating random walks",
    )
    parser.add_argument(
        "--num_walks_per_node", default=1, type=int, help="walks per node"
    )
    parser.add_argument("--walk_type", default="dfs", help="walk type bfs/dfs")
    parser.add_argument("--max_length", default=5, type=int, help="max length of walks")
    parser.add_argument(
        "--n_pred", default=1, help="number of tokens masked to be predicted"
    )
    parser.add_argument(
        "--max_pred", default=1, help="max number of tokens masked to be predicted"
    )
    # Pretraining options
    parser.add_argument("--lr", default=0.0001, type=float, help="learning rate")
    parser.add_argument(
        "--n_epochs", default=50, type=int, help="number of epochs for training"
    )
    parser.add_argument(
        "--checkpoint", default=100, type=int, help="checkpoint for validation"
    )
    parser.add_argument(
        "--base_embedding_dim",
        default=200,
        type=int,
        help="dimension of base embedding",
    )
    parser.add_argument(
        "--batch_size",
        default=1024,
        type=int,
        help="number of data sample in each batch",
    )
    parser.add_argument(
        "--emb_dir",
        default="data",
        type=str,
        help="Used to generate embeddings path if --pretrained_embeddings is not set",
    )
    parser.add_argument(
        "--get_bert_encoder_embeddings",
        default=False,
        help="indicate if need to get node vectors from BERT encoder output, save code "
             "commented out in src/pretraining.py",
    )
    # BERT Layer options
    parser.add_argument(
        "--n_layers", default=4, type=int, help="number of encoder layers in bert"
    )
    parser.add_argument(
        "--d_model", default=200, type=int, help="embedding size in bert"
    )
    parser.add_argument("--d_k", default=64, type=int, help="dimension of K(=Q), V")
    parser.add_argument("--d_v", default=64, type=int, help="dimension of K(=Q), V")
    parser.add_argument("--n_heads", default=4, type=int, help="number of head in bert")
    parser.add_argument(
        "--d_ff",
        default=200 * 4,
        type=int,
        help="4*d_model, FeedForward dimension in bert",
    )
    # GCN Layer options
    parser.add_argument(
        "--is_pre_trained",default=False,
        action="store_true",
        help="if there is pretrained node embeddings",
    )
    parser.add_argument(
        "--gcn_option",
        default="no_gcn",
        help="preprocess bert input once or alternate gcn and bert, preprocess|alternate|no_gcn",
    )
    parser.add_argument(
        "--num_gcn_layers", default=4, type=int, help="number of gcn layers before bert"
    )
    parser.add_argument(
        "--node_edge_composition_func",
        default="no_rel",
        help="options for node and edge compostion, sub|circ_conv|mult|no_rel",
    )

    # Finetuning options
    parser.add_argument("--ft_lr", default=0.0001, type=float, help="learning rate")
    parser.add_argument(
        "--ft_batch_size",
        default=1024,
        type=int,
        help="number of data sample in each batch",
    )
    parser.add_argument(
        "--ft_checkpoint", default=1000, type=int, help="checkpoint for validation"
    )
    parser.add_argument(
        "--ft_d_ff", default=512, type=int, help="feedforward dimension in finetuning"
    )
    parser.add_argument(
        "--ft_layer", default="ffn", help="options for finetune layer: linear|ffn"
    )
    parser.add_argument(
        "--ft_drop_rate", default=0.1, type=float, help="dropout rate in finetuning"
    )
    parser.add_argument(
        "--ft_input_option",
        default="last4_cat",
        help="which output layer from graphbert will be used for finetuning, last|last4_cat|last4_sum",
    )
    parser.add_argument(
        "--false_edge_gen",
        default="basic",
        help="false edge generation pattern/double/basic",
    )
    parser.add_argument(
        "--ft_n_epochs", default=10, type=int, help="number of epochs for training"
    )
    parser.add_argument(
        "--path_option",
        default="shortest",
        help="fine tuning context generation: shortest/all/pattern/random",
    )
    args = parser.parse_args(args=[])

    # Default values
    args.pretrained_embeddings, args.base_embedding_dim = get_set_embeddings_details(
        args
    )
    args.d_model = args.base_embedding_dim
    args.d_ff = args.base_embedding_dim
    args.data_name = data_name
    args.data_path = data_path
    args.outdir = outdir
    args.pretrained_embeddings = pretrained_embeddings
    #args.n_epochs = n_epochs
    #args.ft_n_epochs = ft_n_epochs

    data_path = f"{args.data_path}/{args.data_name}"
    attr_graph, context_gen = get_graph(args, data_path, args.false_edge_gen)
    attr_graph.dump_stats()
    stime = time.time()
    id_maps_dir = data_path
    ent2id = get_id_map(f"{id_maps_dir}/ent2id.txt")
    rel2id = get_id_map(f"{id_maps_dir}/rel2id.txt")

    pretrained_node_embedding = load_pretrained_node2vec(
            args.pretrained_embeddings, ent2id, args.base_embedding_dim
        )
    outdir = data_path+"/outdir"
    isExists = os.path.exists(outdir)
    if isExists:
        shutil.rmtree(outdir)
        os.makedirs(outdir)
    if not isExists:
        os.makedirs(outdir)
    
    outpath = args.outdir + "/"
    isExists = os.path.exists(outpath)
    if isExists:
        shutil.rmtree(outpath)
        os.makedirs(outpath)
    if not isExists:
        os.makedirs(outpath)
    
    finetune_path = outpath + "/finetune/"
    try:
        shutil.rmtree(finetune_path)
        os.mkdir(finetune_path)
    except:  # FIXME - need to replace bare except
        os.mkdir(finetune_path)
    ft_out_dir = finetune_path + "results"
    try:
        shutil.rmtree(ft_out_dir)
        os.mkdir(ft_out_dir)
    except:  # FIXME - need to replace bare except
        os.mkdir(ft_out_dir)
    
    test_path = data_path + "/test.txt"
    test_edges_paths = [test_path]
    test_edges = list(get_test_edges(test_edges_paths, " "))
    #print("No. edges in test data: ", len(test_edges))

    pre_num_batches = setup_pretraining_input_pre(args, attr_graph, context_gen, args.outdir)
    ft_num_batches = setup_finetuning_input_pre(args, attr_graph, context_gen)

    (
        pred_data_test,
        true_data_test,
    ) = Predict(
        args, attr_graph, ft_num_batches, pretrained_node_embedding, ent2id, rel2id, model_path,best_id,is_trained,model_name
    )

    y_pred = np.zeros(len(pred_data_test), dtype=np.int32)

    for i, _ in enumerate(pred_data_test):
        if pred_data_test[i] >= 0.5:
            y_pred[i] = 1
    print('Done')
    CCIlist_new = np.loadtxt(test_path,skiprows=0,dtype=str,delimiter=' ')
    columns = [0, 1, 2]
    CCIlist_new_score = CCIlist_new[:, columns]
    CCIlist_new_score = np.hstack((CCIlist_new_score, y_pred[:, np.newaxis]))
    np.savetxt(data_path + '/predict_ccc.txt', CCIlist_new_score, fmt='%s', delimiter=' ')
    #os.remove(outdir + '/test.txt')
    #return y_pred


import random


class GenerateGraphMask:
    """
    Given a list of subgraphs , where each subgraph is a list of edges, where
    each
    edge.tensor = cat(src_node.tensor, dest_node.tensor, relation.tensor)
    Provides fuctions to generate:
        link preiction task for
        1) subgraph edge prediction
        2) edge.source prediction
        3) edge.relation prediction
        4) edge.target prediction
    """

    def __init__(self, subgraph_sequences, n_pred):
        self.subgraph_sequences = subgraph_sequences
        self.n_pred = n_pred

    def relation_prediction(self):
        """
        generate relation mask
        """
        mask = []
        for subgraph in self.subgraph_sequences:
            num_edges = len(subgraph)
            # print(num_edges, self.n_pred)
            edge_mask_index = random.sample(range(num_edges), self.n_pred)
            subgraph_mask = []
            for i in range(len(subgraph)):
                if i in edge_mask_index:
                    subgraph_mask.append((1, 1, 0))
                else:
                    subgraph_mask.append((1, 1, 1))
            mask.append(subgraph_mask)
        return mask

    def source_prediction(self):
        """
        generate source mask
        """
        mask = []
        for subgraph in self.subgraph_sequences:
            num_edges = len(subgraph)
            edge_mask_index = random.sample(range(num_edges), self.n_pred)
            subgraph_mask = []
            for i in range(len(subgraph)):
                if i in edge_mask_index:
                    subgraph_mask.append((0, 1, 1))
                else:
                    subgraph_mask.append((1, 1, 1))
            mask.append(subgraph_mask)
        return mask

    def target_prediction(self):
        """
        generate target mask
        """
        mask = []
        for subgraph in self.subgraph_sequences:
            num_edges = len(subgraph)
            edge_mask_index = random.sample(range(num_edges), self.n_pred)
            subgraph_mask = []
            for i in range(len(subgraph)):
                if i in edge_mask_index:
                    subgraph_mask.append((1, 0, 1))
                else:
                    subgraph_mask.append((1, 1, 1))
            mask.append(subgraph_mask)
        return mask

    def MaskGeneration(self, mask_type):
        if mask_type == "relation":
            input_mask = self.relation_prediction()
        elif mask_type == "source":
            input_mask = self.source_prediction()
        elif mask_type == "target":
            input_mask = self.target_prediction()
        else:
            print("\n Wrong mask type. No mask is generated!")
            input_mask = []
            for subgraph in self.subgraph_sequences:
                subgraph_mask = [(1, 1, 1) for i in range(len(subgraph))]
                input_mask.append(subgraph_mask)

        return input_mask

    def GCN_MaskGeneration(self):
        mask = []
        masked_nodes = []
        masked_position = []
        for subgraph in self.subgraph_sequences:
            num_nodes = len(subgraph)
            mask_index = random.sample(range(num_nodes), self.n_pred)
            subgraph_mask = []
            subgraph_masked_nodes = []
            subgraph_masked_position = []
            for i, _ in enumerate(subgraph):
                if i in mask_index:
                    subgraph_mask.append(0)
                    subgraph_masked_nodes.append(subgraph[i])
                    subgraph_masked_position.append(i)
                else:
                    subgraph_mask.append(1)
            mask.append(subgraph_mask)
            masked_nodes.append(subgraph_masked_nodes)
            masked_position.append(subgraph_masked_position)

        return mask, masked_nodes, masked_position
